class Operator:
    """
    Default class for creating custom operators. The class can either be extended or a custom function can be given.

    Args:
        operator_label (str): The label used for node representation (default = '').
        function (Callable): A custom function used for performing the execution (default = None).
        min_arity (int): The minimum arity of the function (default = 0).
        max_arity (int): The maximum arity of the function (default = 10).
        strict_precedence (bool): Whether the order of the arguments are strictly required (default = True).
    """

    # TODO define strongly typed sets

    def __init__(self, operator_label : str = '', function = None, min_arity : int = 0, max_arity : int = 10, strict_precedence : bool = True):
        self.operator_label = operator_label
        self.function = function
        self.max_arity = max_arity
        self.min_arity = min_arity
        self.strict_precedence = strict_precedence

    def exec(self, args):
        """
        Takes a list of values (arguments) and performs the given or defined function, then returns the result.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            result: The resulting value.
        """
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")
        return self.function(args)

class OperatorReturn(Operator):

    def __init__(self, operator_label: str = 'return'):
        """
        Merely returns the given argument.

        Args:
            operator_label (str): The label used for node representation (default = 'return').
        """
        super().__init__(operator_label=operator_label, min_arity=1, max_arity=1)

    def exec(self, args):
        """
        Takes a list of length 1 and returns the value.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            result: The resulting value.
        """
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")
        return args[0]

class OperatorAdd(Operator):
    """
    Simple addition of arguments.

    Args:
        operator_label (str): The label used for node representation (default = 'add').
        max_arity (int): The maximum arity of the function (default = 2).
    """
    def __init__(self, operator_label: str = 'add', max_arity : int = 2):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=max_arity, strict_precedence=False)

    def exec(self, args):
        """
        Takes a list of values (arguments) and performs addition, then returns the result.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            result: The resulting value.
        """
        if len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")
        a = 0
        for v in args:
            a += v
        return a

class OperatorSub(Operator):
    """
    Simple subtraction of arguments.

    Args:
        operator_label (str): The label used for node representation (default = 'sub').
        max_arity (int): The maximum arity of the function (default = 2).
    """
    def __init__(self, operator_label: str = 'sub', max_arity : int = 2):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=max_arity, strict_precedence=True)

    def exec(self, args):
        """
        Takes a list of values (arguments) and performs subtraction, then returns the result.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            result: The resulting value.
        """
        if len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")
        a = args[0]
        for v in args[1:]:
            a -= v
        return a

class OperatorMul(Operator):
    """
    Simple multiplication of arguments.

    Args:
        operator_label (str): The label used for node representation (default = 'mul').
        max_arity (int): The maximum arity of the function (default = 2).
    """
    def __init__(self, operator_label: str = 'mul', max_arity : int = 2):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=max_arity, strict_precedence=False)

    def exec(self, args):
        """
        Takes a list of values (arguments) and performs multiplication, then returns the result.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            result: The resulting value.
        """
        if len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")
        a = args[0]
        for v in args[1:]:
            a = a * v
        return a

class OperatorDiv(Operator):
    """
    Simple division of arguments.

    Args:
        operator_label (str): The label used for node representation (default = 'div').
        max_arity (int): The maximum arity of the function (default = 2).
    """
    def __init__(self, operator_label: str = 'div', max_arity : int = 2):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=max_arity, strict_precedence=True)

    def exec(self, args):
        """
        Takes a list of values (arguments) and performs division, then returns the result.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            result: The resulting value.
        """
        if len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")
        a = args[0]
        for v in args[1:]:
            a = a / v
        return a

class OperatorPow(Operator):
    """
    Simple power of arguments.

    Args:
        operator_label (str): The label used for node representation (default = 'pow').
        max_arity (int): The maximum arity of the function (default = 2).
    """
    def __init__(self, operator_label: str = 'pow', max_arity : int = 2):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=max_arity, strict_precedence=True)

    def exec(self, args):
        """
        Takes a list of values (arguments) and performs powering, then returns the result.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            result: The resulting value.
        """
        if len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")
        a = args[0]
        for v in args[1:]:
            a = a ** v
        return a

class OperatorLTE(Operator):
    """
    Simple logical operation, less than or equal.

    Args:
        operator_label (str): The label used for node representation (default = 'lte').
    """
    def __init__(self, operator_label: str = 'lte'):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=2, strict_precedence=True)

    def exec(self, args):
        """
        Takes a list of values (arguments) and tests if the left most is less than or equal to the right.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            bool: The resulting value.
        """
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] <= args[1]

class OperatorLT(Operator):
    """
    Simple logical operation, less than.

    Args:
        operator_label (str): The label used for node representation (default = 'lt').
    """
    def __init__(self, operator_label: str = 'lt'):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=2, strict_precedence=True)

    def exec(self, args):
        """
        Takes a list of values (arguments) and tests if the left most is less than to the right.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            bool: The resulting value.
        """
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] < args[1]

class OperatorGTE(Operator):
    """
    Simple logical operation, greater than or equal.

    Args:
        operator_label (str): The label used for node representation (default = 'gte').
    """
    def __init__(self, operator_label: str = 'gte'):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=2, strict_precedence=True)

    def exec(self, args):
        """
        Takes a list of values (arguments) and tests if the left most is greater than or equal to the right.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            bool: The resulting value.
        """
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] >= args[1]

class OperatorGT(Operator):
    """
    Simple logical operation, greater than or equal.

    Args:
        operator_label (str): The label used for node representation (default = 'gt').
    """
    def __init__(self, operator_label: str = 'gt'):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=2, strict_precedence=True)

    def exec(self, args):
        """
        Takes a list of values (arguments) and tests if the left most is greater than to the right.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            bool: The resulting value.
        """
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] > args[1]

class OperatorEq(Operator):
    """
    Simple logical operation, arguments are equal.

    Args:
        operator_label (str): The label used for node representation (default = 'eq').
    """
    def __init__(self, operator_label: str = 'eq'):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=2, strict_precedence=False)

    def exec(self, args):
        """
        Takes a list of values (arguments) and tests if the left most is equal to the right.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            bool: The resulting value.
        """
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] == args[1]

class OperatorNE(Operator):
    """
    Simple logical operation, arguments are not equal.

    Args:
        operator_label (str): The label used for node representation (default = 'ne').
    """
    def __init__(self, operator_label: str = 'ne'):
        super().__init__(operator_label=operator_label, min_arity=2, max_arity=2, strict_precedence=False)

    def exec(self, args):
        """
        Takes a list of values (arguments) and tests if the left most is not equal to the right.

        Args:
            args (list): The list of ordered arguments for the function.

        Raises:
            AssertionError: If the length of the list of args is not in the range of the min and max arity.

        Returns:
            bool: The resulting value.
        """
        if len(args) > self.max_arity or len(args) < self.min_arity:
            raise AssertionError(
                f"Number of arguments do not match the allowed arity min:{self.min_arity} max:{self.max_arity}")

        return args[0] != args[1]


import random

def crossover_two_one_point(individuals: list, island=None) -> list:
    """
    Classic One Point crossover.

    Args:
        individuals (list): A list (length of 2) of Programs to perform crossover.
        island (Island): The Island calling the method (default = None).

    Returns:
        list: Two new Individuals.
    """
    assert len(individuals) > 1, "Not enough individuals given!"
    mother = individuals[0].node_tree
    father = individuals[1].node_tree

    mother_depth = mother.depth()
    father_depth = father.depth()

    assert mother_depth > 1, f"{individuals[0].name} has a depth of {mother_depth}"
    assert father_depth > 1, f"{individuals[1].name} has a depth of {father_depth}"

    size = min(mother_depth, father_depth)

    point = random.randint(2, size - 1) if size > 2 else 2

    mother_breadth = mother.breadth(depth=point)
    father_breadth = father.breadth(depth=point)

    random_point_mother = random.randint(0, mother_breadth - 1)
    random_point_father = random.randint(0, father_breadth - 1)

    mother_sub_tree = mother.get_subtree(depth=point, index=random_point_mother)

    father_sub_tree = father.get_subtree(depth=point, index=random_point_father)

    mother.set_subtree(depth=point, index=random_point_mother, subtree=father_sub_tree)

    father.set_subtree(depth=point, index=random_point_father, subtree=mother_sub_tree)

    return [individuals[0], individuals[1]]


# -*- coding: utf-8 -*-
"""Basic classes for running Genetic Algorithms.
"""
__author__ = "Justin Hocking"
__copyright__ = "Copyright 2021, Zipfian Science"
__credits__ = []
__license__ = ""
__version__ = "0.0.1"
__maintainer__ = "Justin Hocking"
__email__ = "justin.hocking@zipfian.science"
__status__ = "Development"

import uuid
from typing import List, Union, Any, Callable, Type
import pickle
import warnings as w
from inspect import isclass

import numpy.random as random
import random as pyrand

import natural_selection.genetic_programs.node_operators as op
from natural_selection.genetic_programs.utils import GeneticProgramError


def random_generate(operators : list = None,
                    terminals : list = None,
                    max_depth : int = None,
                    min_depth : int = None,
                    growth_mode : str = 'grow',
                    terminal_prob : float = None,
                    genetic_program = None):
    """
    Function for random node tree creation. This function can be used for either Full tree generation or Grow generation.
    Additionally, minimum and maximum depths along with the probability of adding a terminal to a node can be defined.

    Args:

        operators (list): List of all operators that nodes can be constructed from (default = None).
        terminals (list): List of all terminals that can be included in the node tree, can be numeric or strings for variables (default = None).
        max_depth (int): Maximum depth that node tree can grow (default = 3).
        min_depth (int): Minimum depth that node tree must be (default = 1).
        growth_mode (str): Type of tree growth method to use, "full" or "grow" (default = "grow").
        terminal_prob (float): Probability of a generated node is a terminal (default = 0.5).
        genetic_program (GeneticProgram): An optional program (default = None).

    Returns:
        Node: The node tree generated.
    """
    current_depth = 1

    if max_depth is None:
        if genetic_program:
            max_depth = genetic_program.max_depth
        else:
            raise GeneticProgramError('No max_depth or genetic program given')

    if min_depth is None:
        if genetic_program:
            min_depth = genetic_program.min_depth
        else:
            raise GeneticProgramError('No max_depth or genetic program given')

    def create_random_node_or_terminal():
        if terminal_prob:
            prob = terminal_prob
        else:
            prob = genetic_program.terminal_prob
        if random.random() > prob:
            return create_random_node()
        return create_random_terminal()

    def create_random_terminal():
        if terminals:
            terminal = pyrand.choice(terminals)
        elif genetic_program:
            terminal = pyrand.choice(genetic_program.terminals)
        else:
            raise GeneticProgramError('No terminals were given or define')
        if isinstance(terminal, str):
            return Node(label=terminal, is_terminal=True)
        else:
            return Node(is_terminal=True, terminal_value=terminal)

    def create_random_node():
        if operators:
            starting_operator = pyrand.choice(operators)
        elif genetic_program:
            starting_operator = pyrand.choice(genetic_program.operators)
        else:
            raise GeneticProgramError('No operators were given or define')
        if isinstance(starting_operator, op.Operator):
            initialised_operator = starting_operator
        elif isclass(starting_operator):
            initialised_operator = starting_operator()
        else:
            raise GeneticProgramError(f'Could not recognise the given operator {repr(starting_operator)}')

        if initialised_operator.min_arity == initialised_operator.max_arity:
            starting_arity = initialised_operator.max_arity
        else:
            starting_arity = random.randint(initialised_operator.min_arity, initialised_operator.max_arity)

        return Node(arity=starting_arity, operator=initialised_operator)

    def add_random_children(node, current_depth, max_depth, min_depth):
        at_least_one_node = False
        for i in range(0, node.arity):
            if current_depth == max_depth-1:
                child = create_random_terminal()
            elif growth_mode == 'full':
                child = create_random_node()
            elif current_depth < min_depth and i == node.arity-1 and not at_least_one_node:
                child = create_random_node()
            else:
                child = create_random_node_or_terminal()
            if not child.is_terminal:
                child = add_random_children(child, current_depth + 1, max_depth, min_depth)
                at_least_one_node = True

            node[i] = child

        return node

    if max_depth == 1:
        return create_random_terminal()

    init_node = create_random_node()

    init_node = add_random_children(init_node, current_depth, max_depth, min_depth)

    return init_node


IS_DETERMINISTIC = False
MEMOISE_CACHE = dict()

class Node:
    """
    Basic class for building Node trees. A node can be either a terminal or a parent.
    When initialised as a terminal node, `is_terminal` has to be set to True and either a `label` or a `terminal_value` has to be set.
    When setting a `terminal_value`, the terminal is a literal, constant value.

    Example: n = Node(is_terminal=True, terminal_value=42).

    On only setting a `label`, the terminal is treated as a variable passed on through the function.

    Example: n = Node(is_terminal=True, label='x').

    Setting the arity is optional for when no children nodes are added.

    Args:
        label (str): Optionally set the label, only used for variable terminals (default = None).
        arity (int): Optionally set the function arity, the norm being 2 for functions (default = 1).
        operator (Operator): If the node is a function, set the operator (default = None).
        is_terminal (bool): Explicitly define if the node is a terminal (default = None).
        terminal_value (Any): Only set if the node is terminal and a constant value (default = None).
        children (list): Add a list of child nodes, list length must match arity (default = None).
    """

    def __init__(self, label : str = None,
                 arity : int = 1,
                 operator : op.Operator = None,
                 is_terminal = False,
                 terminal_value = None,
                 children : List = None):
        if label:
            self.label = label
        elif operator:
            self.label = operator.operator_label
        else:
            self.label = str(terminal_value)
        self.arity = 1 if is_terminal else arity
        self.operator = operator
        self.is_terminal = is_terminal
        self.terminal_value = terminal_value
        if children:
            self.children = children
            self.arity = len(children)
        elif not is_terminal:
            self.children = [None] * self.arity
        else:
            self.children = list()

    def __call__(self, **kwargs):

        if self.is_terminal:
            if self.label in kwargs.keys():
                return kwargs[self.label]
            return self.terminal_value
        else:
            if IS_DETERMINISTIC:
                rep = f"{repr(self)} {repr(kwargs)}"
                if rep in MEMOISE_CACHE:
                    return MEMOISE_CACHE[rep]

                result = self.operator.exec([x(**kwargs) for x in self.children])
                MEMOISE_CACHE[rep] = result
                return result

            return self.operator.exec([x(**kwargs) for x in self.children])

    def __str__(self):
        """
        Essentially equivalent to __repr__, but returns a string in the natural order of nodes.
        Where two functionally same trees will return different string representations. Use __repr__ when comparing tree strings.

        Returns:
            str: String representation of tree in natural order of symbols/labels.
        """
        if self.is_terminal:
            return self.label
        else:
            return f"{self.label}({', '.join([str(x) for x in self.children])})"

    def __repr__(self):
        """
        Essentially equivalent to __str__, but more precisely returns an alphabetically sorted str.
        Where two functionally same trees might return different string representations, they will have the exact same __repr__ string.

        Returns:
            str: String representation of tree in alphabetic order of symbols/labels.
        """
        if self.is_terminal:
            return self.label
        else:
            labels = list()
            for n in self.children:
                labels.append(repr(n))
            if self.operator and self.operator.strict_precedence:
                return f"{self.label}({', '.join(labels)})"
            else:
                return f"{self.label}({', '.join(sorted(labels))})"

    def __setitem__(self, index, node):
        if isinstance(index, slice):
            assert index.start < len(self.children), 'Index Out of bounds!'
        else:
            assert index < len(self.children), 'Index Out of bounds!'

        assert isinstance(node, Node), 'Child must be of type Node'

        self.children[index] = node

    def __getitem__(self, index):
        if isinstance(index, slice):
            assert index.start < len(self.children), 'Index Out of bounds!'
        else:
            assert index < len(self.children), 'Index Out of bounds!'

        return self.children[index]

    def __iter__(self):
        self.__n = 0
        return self

    def __next__(self):
        if self.__n < len(self.children):
            gene = self.children[self.__n]
            self.__n += 1
            return gene
        else:
            raise StopIteration

    def __len__(self):
        return len(self.children)

    def clear_children(self):
        self.children = [None] * self.arity

    def is_empty(self):
        for child in self.children:
            if not child is None:
                return False
        return True

    def depth(self):
        """
        Finds the depth of the current tree. This is done by traversing the tree and returning the deepest depth found.

        Returns:
            int: Deepest node depth found.
        """
        if self.is_terminal:
            return 1
        deepest = 0
        for n in self.children:
            if n is None:
                raise GeneticProgramError("A child node is of NoneType")
            d = n.depth() + 1
            if d > deepest:
                deepest = d
        return deepest

    def __breadth(self, to_depth : int , current_depth : int = 0):

        if current_depth == to_depth-1:
            return len(self.children)
        else:
            b = 0
            for n in self.children:
                b += n.__breadth(to_depth=to_depth, current_depth=current_depth+1)
            return b

    def breadth(self, depth: int = None):
        """
        Find the breadth at a given depth of the tree. If no depth is given (None), default to the trees depth.

        Args:
            depth (int): At which depth to find the breadth (default = None).

        Raises:
            GeneticProgramError: If the depth > tree depth or depth < 0.

        Returns:
            int: The breadth of the given depth.
        """
        if depth is None:
            depth = self.depth()
        if depth > self.depth() or depth < 1:
            raise GeneticProgramError(f'Current tree depth not reachable at {depth}')
        return self.__breadth(to_depth=depth, current_depth=1)

    def max_breadth(self):
        """
        Find the broadest part of the tree, returning the breadth and depth at which it occurs.

        Returns:
            tuple: Breadth, depth
        """
        depth = self.depth()
        max_breadth = 0
        max_breadth_depth = 0

        for i in range(depth):
            b_at = self.breadth(i+1)
            if b_at > max_breadth:
                max_breadth_depth = i+1
                max_breadth = b_at

        return max_breadth, max_breadth_depth

    def __get_tree(self, current_depth : int, to_depth : int, ids_only : bool = False):
        if current_depth == to_depth-1:
            if ids_only:
                return [hex(id(n)) for n in self.children]
            else:
                return self.children
        else:
            kids = list()
            for n in self.children:
                kids.extend(n.__get_tree(to_depth=to_depth, current_depth=current_depth+1, ids_only=ids_only))
            return kids

    def get_subtree(self, depth : int, index : int):
        """
        Gets a subtree at the given depth and zero-indexed point.

        Args:
            depth (int): The depth at which to find the subtree.
            index (int): The zero-indexed point at the given depth.

        Raises:
            GeneticProgramError: If the depth > tree depth or depth < 0, or index > length of nodes.

        Returns:
            Node: The subtree at the point.
        """
        if depth > self.depth() or depth < 1:
            raise GeneticProgramError(f'Current tree depth not reachable at {depth}')

        child_nodes = self.__get_tree(current_depth=1, to_depth=depth)

        if index >= len(child_nodes):
            raise GeneticProgramError(f'Index {index} out of range for nodes at depth {depth}')

        return child_nodes[index]

    def __set_tree(self, current_depth : int, to_depth : int, tree, tree_id):
        if current_depth == to_depth-1:
            for i in range(len(self.children)):
                if hex(id(self.children[i])) == tree_id:
                    self.children[i] = tree
                    return True
            return False
        else:
            for n in self.children:
                if n.__set_tree(to_depth=to_depth, current_depth=current_depth+1, tree=tree, tree_id=tree_id):
                    break

    def set_subtree(self, depth : int, index : int, subtree):
        """
        Set the subtree at the given depth and the index.

        Raises:
            GeneticProgramError: If the depth > tree depth or depth < 0, or index > length of nodes.

        Args:
            depth (int): The depth at which to set the subtree.
            index (int): The zero-indexed point at the given depth to set the subtree.
            subtree (Node): The subtree to set at the given point.

        """
        if depth > self.depth() or depth < 1:
            raise GeneticProgramError(f'Current tree depth not reachable at {depth}')

        ids = self.__get_tree(current_depth=1, to_depth=depth, ids_only=True)

        if index >= len(ids):
            raise GeneticProgramError(f'Index {index} out of range for nodes at depth {depth}')

        id_ = ids[index]

        self.__set_tree(current_depth=1, to_depth=depth, tree=subtree, tree_id=id_)



class GeneticProgram:
    """
    A class that encapsulates a single genetic program, with node tree and a fitness evaluation function.

    Args:
        fitness_function (Callable): Function with ``func(Node, island, **params)`` signature (default = None).
        node_tree (Node): A starting node tree (default = None).
        operators (list): List of all operators that nodes can be constructed from (default = None).
        terminals (list): List of all terminals that can be included in the node tree, can be numeric or strings for variables (default = None).
        max_depth (int): Maximum depth that node tree can grow (default = 3).
        min_depth (int): Minimum depth that node tree must be (default = 1).
        growth_mode (str): Type of tree growth method to use, "full" or "grow" (default = "grow").
        terminal_prob (float): Probability of a generated node is a terminal (default = 0.5).
        tree_generator (Callable): Function with to create the tree.
        name (str): Name for keeping track of lineage (default = None).
        species_type (str) : A unique string to identify the species type, for preventing cross polluting (default = None).
        filepath (str): Skip init and load from a pickled file.
        is_deterministic (bool): Sets the whole program as deterministic, which improves function calls through memoisation (default = False).
        program_properties (dict): For fitness functions, extra params may be given (default = None).

    Notes:
        If no fitness function is given, the default __call__ is used.

    Attributes:
        fitness (Numeric): The fitness score after evaluation.
        age (int): How many generations was the individual alive.
        genetic_code (str): String representation of node tree.
        history (list): List of dicts of every evaluation.
        parents (list): List of strings of parent names.
    """

    def __init__(self,
                 fitness_function: Callable = None,
                 node_tree : Node = None,
                 operators : List[Union[Type,op.Operator]] = None,
                 terminals : List[Union[str,int,float]]= None,
                 max_depth: int = 3,
                 min_depth : int = 1,
                 growth_mode: str = 'grow',
                 terminal_prob: float = 0.5,
                 tree_generator: Callable = random_generate,
                 name: str = None,
                 species_type : str = None,
                 filepath : str = None,
                 is_deterministic: bool = False,
                 program_properties : dict = None):
        if not filepath is None:
            self.load(filepath=filepath)
            return
        if fitness_function and '<lambda>' in repr(fitness_function):
            w.warn("WARNING: 'fitness_function' lambda can not be pickled using standard libraries.")
        if name is None:
            self.name = str(uuid.uuid4())
        else:
            self.name = name
        self.operators = operators
        self.terminals = terminals
        self.max_depth = max_depth
        self.min_depth = min_depth
        self.node_tree = node_tree if not node_tree is None else tree_generator(
            operators=operators,
            terminals=terminals,
            max_depth=max_depth,
            min_depth=min_depth,
            growth_mode=growth_mode,
            terminal_prob=terminal_prob,
            genetic_program=self
        )
        self.terminal_prob = terminal_prob
        self.growth_mode = growth_mode
        self.tree_generator = tree_generator

        self.root_node = Node(label=self.name,arity=1, operator=op.OperatorReturn())

        self.fitness_function = fitness_function
        self.fitness = None
        self.age = 0
        self.genetic_code = None
        self.history = list()
        self.parents = list()
        if species_type:
            self.species_type = species_type
        else:
            self.species_type = "def"

        self.is_deterministic = is_deterministic
        global IS_DETERMINISTIC
        IS_DETERMINISTIC = is_deterministic

        self.__program_properties = program_properties
        if program_properties:
            for k, v in program_properties.items():
                self.__dict__.update({k: v})

    def __call__(self, **kwargs):
        if not self.node_tree is None:
            self.root_node[0] = self.node_tree
        else:
            raise GeneticProgramError('No node tree has been constructed!')
        return self.root_node(**kwargs)

    def depth(self):
        """
        Wrapping around the depth function of the node tree.

        Returns:
            int: Node tree depth.
        """
        return self.node_tree.depth()

    def register_parent_names(self, parents: list, reset_parent_name_list: bool = True):
        """
        In keeping lineage of family lines, the names of parents are kept track of.

        Args:
            parents (list): A list of GeneticProgram of the parents.
        """
        if reset_parent_name_list:
            self.parents = list()
        for parent in parents:
            self.parents.append(parent)


    def reset_name(self, name: str = None):
        """
        A function to reset the name of a program, helping to keep linage of families.

        Args:
            name (str): Name (default = None).
        """
        if name is None:
            self.name = str(uuid.uuid4())
        else:
            self.name = name


    def birthday(self, add: int = 1):
        """
        Add to the age. This is for keeping track of how many generations a program has "lived" through.

        Args:
            add (int): Amount to age.
        """
        self.age += add


    def reset_fitness(self, fitness: Any = None, reset_genetic_code: bool = True):
        """
        Reset (or set) the fitness of the program.

        Args:
            fitness (Any): New fitness value (default = None).
            reset_genetic_code (bool): Whether to reset the genetic code. (default = True)
        """
        self.fitness = fitness
        if reset_genetic_code:
            self.genetic_code = None

    def evaluate(self, params : dict = None, island=None) -> Any:
        """
        Run the fitness function with the given params.

        Args:
            params (dict): Named dict of eval params.
            island (Island): Pass the Island for advanced fitness functions based on Island properties and populations.

        Returns:
            numeric: Fitness value.
        """
        if not params is None:
            _params = params
        else:
            _params = {}
        try:
            if self.fitness_function is None:
                self.fitness = self(**_params)
            else:
                self.fitness = self.fitness_function(program=self, island=island, **_params)
        except Exception as exc:
            if island:
                island.verbose_logging(f"ERROR: {self.name} - {repr(self.node_tree)} - {repr(exc)}")
            raise GeneticProgramError(message=f'Could not evaluate program "{self.name}" due to {repr(exc)}')

        stamp = { "name": self.name,
                  "age": self.age,
                  "fitness": self.fitness,
                  "node_tree": str(self.node_tree),
                  "parents": self.parents,
         }

        if island:
            stamp["island_generation" ] = island.generation_count

        self.history.append(stamp)
        return self.fitness

    def unique_genetic_code(self, force_update : bool = False) -> str:
        """
        Gets the unique genetic code, generating if it is undefined.

        Args:
            force_update (bool): Force update of genetic_code property (default = False).

        Returns:
            str: String name of Chromosome.
        """
        if self.genetic_code is None or force_update:
            self.genetic_code = repr(self.node_tree)
        return self.genetic_code

    def save(self, filepath : str):
        """
        Save an individual to a pickle file.

        Args:
            filepath (str): File path to write to.
        """
        with open(filepath, "wb") as f:
            pickle.dump(self.__dict__, f)

    def load(self, filepath : str):
        """
        Load an individual from a pickle file.

        Args:
            filepath (str): File path to load from.
        """
        with open(filepath, "rb") as f:
            self.__dict__.update(pickle.load(f))

    def add_new_property(self, key : str, value : Any):
        """
        Method to add new properties (attributes).

        Args:
            key (str): Name of property.
            value (Any): Anything.
        """
        if self.__program_properties:
            self.__program_properties.update({key: value})
        else:
            self.__program_properties = {key: value}
        self.__dict__.update({key: value})

    def get_properties(self) -> dict:
        """
        Gets a dict of the custom properties that were added at initialisation or the `add_new_property` method.

        Returns:
            dict: All custom properties.
        """
        return self.__program_properties

    def get(self, key : str, default=None):
        """
        Gets the value of a property or returns default if it doesn't exist.

        Args:
            key (str): Property name.
            default: Value to return if the property is not found (default = None).

        Returns:
            any: The property of the individual.
        """
        if key in self.__dict__.keys():
            return self.__dict__[key]
        return default

    def __str__(self) -> str:
        return f'GeneticProgram({self.name}:{self.fitness})'

    def __repr__(self) -> str:
        genetic_code = self.unique_genetic_code()
        return f'GeneticProgram({self.name}:{self.fitness}:{self.age}:{self.species_type}:{genetic_code})'

    def __eq__(self, other):
        if isinstance(other, GeneticProgram):
            return self.unique_genetic_code() == other.unique_genetic_code()
        else:
            raise GeneticProgramError(message=f'Can not compare {type(other)}')

    def __ne__(self, other):
        if isinstance(other, GeneticProgram):
            return self.unique_genetic_code() != other.unique_genetic_code()
        else:
            raise GeneticProgramError(message=f'Can not compare {type(other)}')

    def __lt__(self, other):
        if isinstance(other, GeneticProgram):
            return self.fitness < other.fitness
        elif isinstance(other, int):
            return self.fitness < other
        elif isinstance(other, float):
            return self.fitness < other
        else:
            raise GeneticProgramError(message=f'Can not compare {type(other)}')

    def __le__(self, other):
        if isinstance(other, GeneticProgram):
            return self.fitness <= other.fitness
        elif isinstance(other, int):
            return self.fitness <= other
        elif isinstance(other, float):
            return self.fitness <= other
        else:
            raise GeneticProgramError(message=f'Can not compare {type(other)}')

    def __gt__(self, other):
        if isinstance(other, GeneticProgram):
            return self.fitness > other.fitness
        elif isinstance(other, int):
            return self.fitness > other
        elif isinstance(other, float):
            return self.fitness > other
        else:
            raise GeneticProgramError(message=f'Can not compare {type(other)}')

    def __ge__(self, other):
        if isinstance(other, GeneticProgram):
            return self.fitness >= other.fitness
        elif isinstance(other, int):
            return self.fitness >= other
        elif isinstance(other, float):
            return self.fitness >= other
        else:
            raise GeneticProgramError(message=f'Can not compare {type(other)}')


import unittest
from natural_selection.genetic_programs.node_operators import OperatorAdd, OperatorSub, OperatorMul, OperatorDiv, Operator
from natural_selection.genetic_programs.operators.crossover import crossover_two_one_point
from natural_selection.genetic_programs import GeneticProgram

import copy

class TestGPCrossover(unittest.TestCase):

    def test_crossover_two_one_point(self):
        operators = [OperatorAdd(operator_label='+'), OperatorSub(operator_label='-')]
        terminals = ['X', 2]

        gp_full = GeneticProgram(operators=operators, terminals=terminals, max_depth=3, min_depth=2, growth_mode='full')

        gp_grow = GeneticProgram(operators=operators, terminals=terminals, max_depth=3, min_depth=2, growth_mode='grow')

        mother_repr = repr(gp_full.node_tree)
        father_repr = repr(gp_grow.node_tree)

        offspring = crossover_two_one_point([copy.deepcopy(gp_full), copy.deepcopy(gp_grow)])

        offspring_1_repr = repr(offspring[0].node_tree)
        offspring_2_repr = repr(offspring[1].node_tree)

        self.assertNotEqual(mother_repr, offspring_1_repr)
        self.assertNotEqual(father_repr, offspring_2_repr)

        mother_repr_after = repr(gp_full.node_tree)
        father_repr_after = repr(gp_grow.node_tree)

        self.assertEqual(mother_repr, mother_repr_after)
        self.assertEqual(father_repr, father_repr_after)




import unittest
from natural_selection.genetic_algorithms import Gene, Chromosome, Individual
from natural_selection import Island
from natural_selection.genetic_algorithms.utils.random_functions import random_int
from natural_selection.genetic_algorithms.operators.crossover import crossover_two_n_point
from natural_selection.utils.population_growth import population_generational, population_steady_state_remove_oldest, population_steady_state_remove_weakest

class TestSimpleIsland(unittest.TestCase):

    def setUp(self) -> None:
        self.fitness = lambda individual, island, x, y: (individual.chromosome[0].value * x) + (
                    individual.chromosome[1].value * y)
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])
        self.ind = Individual(self.fitness, name="Adam", chromosome=gen)

    def test_population_steady_state_remove_oldest(self):

        life = Island({'x': 0.6, 'y' : 0.2}, crossover_function=crossover_two_n_point,
                      population_growth_function=population_steady_state_remove_oldest,
                      allow_twins=True
                      )
        life.initialise(self.ind, population_size=10)

        best = life.evolve(crossover_params={'n_points' : 2})

        self.assertEqual(len(life.population), 10)

    def test_population_steady_state_remove_weakest(self):

        life = Island({'x': 0.6, 'y' : 0.2}, crossover_function=crossover_two_n_point,
                      population_growth_function=population_steady_state_remove_weakest,
                      allow_twins=True
                      )
        life.initialise(self.ind, population_size=10)

        best = life.evolve(crossover_params={'n_points' : 2})

        self.assertEqual(len(life.population), 10)

    def test_population_generational(self):

        life = Island({'x': 0.6, 'y' : 0.2},
                      crossover_function=crossover_two_n_point,
                      population_growth_function=population_generational,
                      allow_twins=True
                      )
        life.initialise(self.ind, population_size=10)

        best = life.evolve(crossover_params={'n_points' : 2},
                           parent_selection_params={'n' : 10},
                           crossover_probability=1.0,)

        self.assertEqual(len(life.population), 10)



__version__ = '0.2.32'
__date__ = "2023-02-15"

from time import gmtime
import multiprocessing as mp
import random
import logging
from datetime import datetime
import warnings as w
from typing import Callable, Any, Iterable, List, Union, Type
import pickle
import copy

import numpy as np

from natural_selection.utils import get_random_string, clone_classic, default_save_checkpoint_function, \
    evaluate_individuals_sequentially, evaluate_individual_multiproc_wrapper, IslandError
from natural_selection.utils.population_growth import population_incremental

from natural_selection.genetic_algorithms import Gene, Chromosome, Individual
from natural_selection.genetic_programs import Node, GeneticProgram, random_generate

from natural_selection.genetic_algorithms.operators.initialisation import initialise_population_random, alien_spawn_default
from natural_selection.genetic_algorithms.operators.selection import selection_elites_top_n, selection_parents_two, selection_survivors_all
from natural_selection.genetic_algorithms.operators.crossover import crossover_two_uniform
from natural_selection.genetic_algorithms.operators.mutation import mutation_randomize
from natural_selection.utils.probability_functions import crossover_prob_function_classic, mutation_prob_function_classic
from natural_selection.genetic_algorithms.utils import GeneticAlgorithmError

from natural_selection.genetic_programs.operators.initialisation import initialise_population_full_method
from natural_selection.genetic_programs.utils import GeneticProgramError
import natural_selection.genetic_programs.node_operators as gp_func


class Island:
    """
    A simple Island to perform a Genetic Algorithm. By default, the selection, mutation, crossover, and probability functions
    default to the classic functions for Genetic Algorithms.

    Args:
        function_params (dict): The parameters for the fitness function (default = None).
        maximise_function (bool): If True, the fitness value is maximised, False will minimise the function fitness (default = True).
        population_growth_function (Callable): Function that defines what the population growth is, steady, generational etc. (default = None).
        parent_selection (Callable): Function for selecting individuals for crossover (default = None).
        initialisation_function (Callable): A function for randomly creating new individuals from the given adam.
        parent_combination (Callable): Function for combining parents for crossover (default = None).
        crossover_function (Callable): Function for crossover (default = None).
        mutation_function (Callable): Function for mutation (default = None).
        crossover_prob_function (Callable): Random probability function for crossover (default = None).
        mutation_prob_function (Callable): Random probability function for mutation (default = None).
        clone_function (Callable): Function for cloning (default = None).
        survivor_selection_function (Callable): Function for selecting survivors (default = None).
        alien_spawn_function (Callable): Function for spawning new aliens during each generation (default = None).
        random_seed (int): Random seed for random and Numpy generators, set to None for no seed (default = None).
        name (str): General name for island, useful when working with multiple islands (default = None).
        verbose (bool): Print all information (default = None).
        logging_function (Callable): Function for custom message logging, such as server logging (default = None).
        save_checkpoint_function (Callable): Function for custom checkpoint saving (default = None).
        filepath (str): If a filepath is specified, the pickled island is loaded from it, skipping the rest of initialisation (default = None).
        save_checkpoint_level (int): Level of checkpoint saving 0 = none, 1 = per generation, 2 = per evaluation (default = 0).
        core_count (int, float): Number of cores to split evaluation on, for all cores, set -1, use float for fractional (default = 1).
        population_evaluation_function (Callable): Function taking a list of individuals and calling the fitness functions (default = None).
        allow_twins (bool): Only add new offspring to the population if they have a unique chromosome (default = False).

    Attributes:
        unique_genome (list): List of unique chromosomes.
        generation_info (list): List of dicts detailing info for every generation.
        population (list): The full population of members.
        parents (list): All parents selected during the run.
        elites (list): All elites selected during the run.
        mutants (list): All mutants created during the run.
        children (list): All children created during the run.
        generation_count (int): The current generation number.
        checkpoints_dir (str): Directory name of where all checkpoints are saved.
        lineage (dict): A graph of lineage for the whole population.
    """

    __stat_key = 'stat'
    __generation_key = 'generation'
    __pop_len_key = 'pop_len'
    __fitness_mean_key = 'fitness_mean'
    __fitness_std_key = 'fitness_std'
    __fitness_min_key = 'fitness_min'
    __fitness_max_key = 'fitness_max'
    __most_fit_key = 'most_fit'
    __least_fit_key = 'least_fit'


    def __init__(self, function_params : dict = None,
                 maximise_function : bool = True,
                 population_growth_function : Callable = population_incremental,
                 initialisation_function: Callable = initialise_population_random,
                 parent_selection : Callable = selection_elites_top_n,
                 parent_combination : Callable = selection_parents_two,
                 crossover_function : Callable = crossover_two_uniform,
                 mutation_function : Callable = mutation_randomize,
                 crossover_prob_function : Callable = crossover_prob_function_classic,
                 mutation_prob_function : Callable = mutation_prob_function_classic,
                 survivor_selection_function: Callable = selection_survivors_all,
                 alien_spawn_function: Callable = alien_spawn_default,
                 clone_function : Callable = clone_classic,
                 random_seed: int = None,
                 name : str = None,
                 verbose : bool = True,
                 logging_function : Callable = None,
                 save_checkpoint_function: Callable = default_save_checkpoint_function,
                 filepath : str = None,
                 save_checkpoint_level : int = 0,
                 core_count : Union[int, float] = 1,
                 population_evaluation_function : Callable = evaluate_individuals_sequentially,
                 allow_twins : bool = False):

        if name is None:
            self.name = get_random_string(include_numeric=True)
        else:
            self.name = name

        self.verbose = verbose
        self.logging_function = logging_function

        if verbose:
            logging.basicConfig(level=logging.INFO,
                                format='%(asctime)s %(island)-8s %(message)s',
                                filename=datetime.utcnow().strftime('%Y-%m-%d-ga-output.log'),
                                datefmt='%H:%M:%S')

            logging.Formatter.converter = gmtime

        if filepath:
            self.load_island(filepath)
            return

        self.verbose_logging(f"island: create v{__version__}")
        if function_params:
            self.verbose_logging(f"island: param {function_params}")
            self.function_params = function_params
        else:
            self.function_params = dict()
        self.unique_genome = list()
        self.generation_info = list()
        self.lineage = {
            'nodes' : [],
            'edges' : []
        }
        self.population = list()

        self.maximise_function = maximise_function
        self.verbose_logging(f"island: maximise_function {maximise_function}")
        self.population_growth_function = population_growth_function
        self.verbose_logging(f"island: population_growth_function {population_growth_function.__name__}")
        self._initialise = initialisation_function
        self.verbose_logging(f"island: initialisation_function {initialisation_function.__name__}")
        self.parent_selection = parent_selection
        self.verbose_logging(f"island: parent_selection {parent_selection.__name__}")
        self.parent_combination = parent_combination
        self.verbose_logging(f"island: parent_combination {parent_combination.__name__}")
        self.crossover = crossover_function
        self.verbose_logging(f"island: crossover_function {crossover_function.__name__}")
        self.mutation = mutation_function
        self.verbose_logging(f"island: mutation_function {mutation_function.__name__}")
        self.crossover_prob = crossover_prob_function
        self.verbose_logging(f"island: crossover_prob_function {crossover_prob_function.__name__}")
        self.mutation_prob = mutation_prob_function
        self.verbose_logging(f"island: mutation_prob_function {mutation_prob_function.__name__}")
        self.clone = clone_function
        self.verbose_logging(f"island: clone_function {clone_function.__name__}")
        self.survivor_selection = survivor_selection_function
        self.verbose_logging(f"island: survivor_selection_function {survivor_selection_function.__name__}")

        self.alien_spawn_function = alien_spawn_function
        self.verbose_logging(f"island: alien_spawn_function {alien_spawn_function.__name__}")

        self.save_checkpoint = save_checkpoint_function
        self.verbose_logging(f"island: save_checkpoint_function {save_checkpoint_function.__name__}")

        self.allow_twins = allow_twins
        self.verbose_logging(f"island: allow_twins {allow_twins}")

        self.random_seed = random_seed
        self.verbose_logging(f"island: random_seed {random_seed}")

        self.population_evaluation_function = population_evaluation_function
        self.verbose_logging(f"island: population_evaluation_function {population_evaluation_function.__name__}")

        if isinstance(core_count, int):
            if core_count < 1:
                self.core_count = mp.cpu_count()
            else:
                self.core_count = core_count
        elif isinstance(core_count, float) and core_count < 1.0 and core_count > 0.0:
            self.core_count = round(core_count * mp.cpu_count())
        else:
            raise ValueError('The value of core_count must be of type float or int, and 0 < core_count < 1 if float')

        self.verbose_logging(f"island: core_count {core_count} as {self.core_count}")

        self.elites = list()
        self.parents = list()
        self.mutants = list()
        self.children = list()
        self.deaths = dict()
        self.species_type = "def"
        self.generation_count = 0
        self.save_checkpoint_level = save_checkpoint_level
        self.checkpoints_dir = f'_ga_checkpoints'
        self.verbose_logging(f"island: save_checkpoint_level {save_checkpoint_level}")

        # Set python random seed, as well as Numpy seed.
        if random_seed:
            random.seed(random_seed)
            from numpy.random import seed as np_seed
            np_seed(random_seed)


        if '<lambda>' in repr(initialisation_function):
            w.warn("WARNING: 'initialisation_function' lambda can not be pickled using standard libraries.")

        if '<lambda>' in repr(parent_selection):
            w.warn("WARNING: 'parent_selection' lambda can not be pickled using standard libraries.")

        if '<lambda>' in repr(parent_combination):
            w.warn("WARNING: 'parent_combination' lambda can not be pickled using standard libraries.")

        if '<lambda>' in repr(crossover_function):
            w.warn("WARNING: 'crossover_function' lambda can not be pickled using standard libraries.")

        if '<lambda>' in repr(mutation_function):
            w.warn("WARNING: 'mutation_function' lambda can not be pickled using standard libraries.")

        if '<lambda>' in repr(crossover_prob_function):
            w.warn("WARNING: 'crossover_prob_function' lambda can not be pickled using standard libraries.")

        if '<lambda>' in repr(mutation_prob_function):
            w.warn("WARNING: 'mutation_prob_function' lambda can not be pickled using standard libraries.")

        if '<lambda>' in repr(clone_function):
            w.warn("WARNING: 'clone_function' lambda can not be pickled using standard libraries.")

        if '<lambda>' in repr(survivor_selection_function):
            w.warn("WARNING: 'survivor_selection_function' lambda can not be pickled using standard libraries.")

        if '<lambda>' in repr(alien_spawn_function):
            w.warn("WARNING: 'alien_spawn_function' lambda can not be pickled using standard libraries.")

        if logging_function and '<lambda>' in repr(logging_function):
            w.warn("WARNING: 'logging_function' lambda can not be pickled using standard libraries.")

        if save_checkpoint_function and '<lambda>' in repr(save_checkpoint_function):
            w.warn("WARNING: 'save_checkpoint_function' lambda can not be pickled using standard libraries.")


        if population_evaluation_function and '<lambda>' in repr(population_evaluation_function):
            w.warn("WARNING: 'population_evaluation_function' lambda can not be pickled using standard libraries.")

        if population_growth_function and '<lambda>' in repr(population_growth_function):
            w.warn("WARNING: 'population_growth_function' lambda can not be pickled using standard libraries.")


        self.__island_properties = dict()

    def create_node(self,
                    label: str = None,
                    arity: int = 1,
                    operator: gp_func.Operator = None,
                    is_terminal=False,
                    terminal_value=None,
                    children: List = None):
        """
        Wrapping function to create a new node. Useful when writing new initialisation functions. See Node class.

        Args:
            label (str): Optionally set the label, only used for variable terminals (default = None).
            arity (int): Optionally set the function arity, the norm being 2 for functions (default = 1).
            operator (Operator): If the node is a function, set the operator (default = None).
            is_terminal (bool): Explicitly define if the node is a terminal (default = None).
            terminal_value (Any): Only set if the node is terminal and a constant value (default = None).
            children (list): Add a list of child nodes, list length must match arity (default = None).

        Returns:
            node: A new Node object.
        """
        return Node(label=label,
             arity=arity,
             operator=operator,
             is_terminal=is_terminal,
             terminal_value=terminal_value,
             children=children)

    def create_gene(self,
                 name : str,
                 value : Any,
                 randomise_function: Callable,
                 gene_max : Any = None,
                 gene_min : Any = None,
                 mu : Any = None,
                 sig: Any = None,
                 step_lower_bound : Any = None,
                 step_upper_bound : Any = None,
                 choices : Iterable  = None,
                 gene_properties : dict = None):
        """
        Wrapping function to create a new Gene. Useful when writing new initialisation functions. See Gene class.

        Args:
            name (str): Gene name. The gene name also acts as a compatibility reference.
            value (Any): The value, could be any type.
            randomise_function (Callable): A function to randomise the gene, taking the min and max as input with signature ``func(self)``.
            gene_max (Any, numeric type): Max value for random number generator (default = None).
            gene_min (Any, numeric type): Min value for random number generator (default = None).
            mu (Any, numeric type): Mean value of distribution to sample from (default = None).
            sig (Any, numeric type): Std. Dev. value of distribution to sample from (default = None).
            step_lower_bound (Any, numeric type): For uniform stepping functions, defines lower bound of range (default = None).
            step_upper_bound (Any, numeric type): For uniform stepping functions, defines upper bound of range (default = None).
            choices (Iterable): List of choices, categorical or not, to randomly choose from (default = None).
            gene_properties (dict): For custom random functions, extra params may be given (default = None).
        Returns:
            gene: A new Gene object.
        """
        return Gene(name=name, value=value, randomise_function=randomise_function, gene_max=gene_max, gene_min=gene_min,
                    mu=mu, sig=sig, step_lower_bound=step_lower_bound, step_upper_bound=step_upper_bound, choices=choices, gene_properties=gene_properties)

    def create_chromosome(self,
                          genes: list = None,
                          gene_verify_func : Callable = None,
                          chromosome_properties : dict = None):
        """
        Wrapping function to create a new Chromosome. Useful when writing new initialisation functions. See Chromosome class.

        Args:
            genes (list): list of initialised Gene objects.
            gene_verify_func (Callable): A function to verify gene compatibility `func(gene,loc,chromosome)` (default = None).
            chromosome_properties (dict): For custom functions, extra params may be given (default = None).

        Returns:
            chromosome: A new Chromosome.
        """
        return Chromosome(genes=genes, gene_verify_func=gene_verify_func, chromosome_properties=chromosome_properties)

    def create_genetic_program(self,
                               fitness_function: Callable = None,
                               node_tree: Node = None,
                               operators: List[Union[Type, gp_func.Operator]] = None,
                               terminals: List[Union[str, int, float]] = None,
                               max_depth: int = 3,
                               min_depth: int = 1,
                               growth_mode: str = 'grow',
                               terminal_prob: float = 0.5,
                               tree_generator: Callable = random_generate,
                               name: str = None,
                               species_type: str = None,
                               add_to_population: bool = False,
                               program_properties: dict = None,
                               is_deterministic: bool = False
                               ):
        """
        Wrapping function to create a new GeneticProgram. Useful when writing new initialisation functions. See GeneticProgram class.

        Args:
            fitness_function (Callable): Function with ``func(Node, island, **params)`` signature (default = None).
            node_tree (Node): A starting node tree (default = None).
            operators (list): List of all operators that nodes can be constructed from (default = None).
            terminals (list): List of all terminals that can be included in the node tree, can be numeric or strings for variables (default = None).
            max_depth (int): Maximum depth that node tree can grow (default = 3).
            min_depth (int): Minimum depth that node tree must be (default = 1).
            growth_mode (str): Type of tree growth method to use, "full" or "grow" (default = "grow").
            terminal_prob (float): Probability of a generated node is a terminal (default = 0.5).
            tree_generator (Callable): Function with to create the tree.
            name (str): Name for keeping track of lineage (default = None).
            species_type (str) : A unique string to identify the species type, for preventing cross polluting (default = None).
            add_to_population (bool): Add this new program to the population (default = False).
            program_properties (dict): For fitness functions, extra params may be given (default = None).
            is_deterministic (bool): Sets the whole program as deterministic, which improves function calls through memoisation (default = False).

        Returns:
            program: Newly created GeneticProgram.
        """
        gp = GeneticProgram(fitness_function=fitness_function,
                              node_tree=node_tree,
                              operators=operators,
                              terminals=terminals,
                              max_depth=max_depth,
                              min_depth=min_depth,
                              growth_mode=growth_mode,
                              terminal_prob=terminal_prob,
                              tree_generator=tree_generator,
                              name=name,
                              species_type=species_type,
                              program_properties=program_properties,
                              is_deterministic=is_deterministic,
                              )
        if add_to_population:
            self.population.append(gp)
        return gp


    def create_individual(self,
                          fitness_function : Callable,
                          name : str = None,
                          chromosome: Chromosome = None,
                          species_type : str = None,
                          add_to_population : bool = False,
                          individual_properties : dict = None):
        """
        Wrapping function to create a new Individual. Useful when writing new initialisation functions. See Individual class.

        Args:
            fitness_function (Callable): Function with ``func(Chromosome, island, **params)`` signature
            name (str): Name for keeping track of lineage (default = None).
            chromosome (Chromosome): A Chromosome object, initialised (default = None).
            species_type (str) : A unique string to identify the species type, for preventing cross polluting (default = None).
            add_to_population (bool): Add this new individual to the population (default = False).
            individual_properties (dict): For fitness functions, extra params may be given (default = None).

        Returns:
            individual: A new Individual.
        """
        ind = Individual(fitness_function=fitness_function, name=name, chromosome=chromosome, species_type=species_type,
                         individual_properties=individual_properties)
        if add_to_population:
            self.population.append(ind)
        return ind

    def initialise(self, adam : Union[Individual,GeneticProgram],
               population_size: int = 8,
               initialisation_params : dict = None,
               evaluate_population : bool = True
               ):
        """
        Starts the population by taking an initial individual as template and creating new ones from it.
        Island ``species_type`` is set to adam's species type.

        Args:
            adam (Individual): Individual to clone from.
            population_size (int): Size of population.
            initialisation_params (dict): Custom params for custom initialisation functions (default = None).
            evaluate_population (bool): Evaluate the newly created population (default = True).
        """
        self.species_type = adam.species_type

        if initialisation_params:
            _initialisation_params = initialisation_params
        else:
            _initialisation_params = {}

        self.verbose_logging(f"init: pop_size {population_size}")
        self.verbose_logging(f"init: adam {str(adam)}")
        self.population = self._initialise(adam=adam, n=population_size, island=self, **_initialisation_params)

        if self.save_checkpoint_level == 2:
            self.save_checkpoint(event='init_pre', island=self)

        if evaluate_population:
            for popitem in self.__evaluate_individuals(self.population):
                self.unique_genome.append(popitem.unique_genetic_code())
        if self.save_checkpoint_level == 2:
            self.save_checkpoint(event='init_post', island=self)
        self.verbose_logging("init: complete")

        for p in self.population:
            p.history.append({"init_island": self.name})
            self.__add_to_lineage(p)

    def __add_to_lineage(self, individual : Union[Individual, GeneticProgram]):
        c = str(individual.chromosome) if isinstance(individual,Individual) else str(individual.node_tree)
        node = {
            'name': individual.name,
            'age': individual.age,
            'fitness': individual.fitness,
            'chromosome': c,
            '_individual' : individual
        }
        properties = individual.get_properties()
        if not properties is None:
            node.update(properties)
        self.lineage['nodes'].append(node)

    def import_migrants(self, migrants : list,
                        reset_fitness : bool = False,
                        species_check : bool = True,
                        allow_twins : bool = True):
        """
        Imports a list of individuals, with option to re-evaluate them.
        Skips the individual if the genetic code is already in the population.

        Args:
            migrants (list): List of Individuals.
            reset_fitness (bool): Reset the fitness of new members and evaluate them (default = False).
            species_check (bool): Safely check that imported members are compatible with population  (default = True).
            allow_twins (bool): Only add migrants to the population if they have a unique chromosome (default = True).
        """
        migrants_for_testing = list()
        migrants_for_adding = list()

        for i in migrants:
            self.verbose_logging(f"migration: customs {str(i)}")
            if species_check:
                if i.species_type != self.species_type:
                    continue
            if allow_twins or not i.unique_genetic_code(force_update=True) in self.unique_genome:
                    if i.fitness is None or reset_fitness:
                        migrants_for_testing.append(i)
                    else:
                        migrants_for_adding.append(i)

        migrants_for_testing = self.__evaluate_individuals(migrants_for_testing)
        migrants_for_adding.extend(migrants_for_testing)

        for i in migrants_for_adding:
            self.verbose_logging(f"migration: add {str(i)}")
            i.history.append({"migration_island": self.name})
            self.population.append(i)
            self.unique_genome.append(i.unique_genetic_code())
            self.__add_to_lineage(i)

        self.verbose_logging("migration: imported")


    def evolve(self, starting_generation : int = 0,
               n_generations : int = 5,
               crossover_probability : float = 0.5,
               mutation_probability : float = 0.25,
               crossover_params : dict = None,
               mutation_params : dict = None,
               parent_selection_params : dict = None,
               parent_combination_params: dict = None,
               survivor_selection_params: dict = None,
               population_growth_params: dict = None,
               alien_spawn_params: dict = None,
               criterion_function : Callable = None,
               criterion_params : dict = None,
               pre_generation_check_function : Callable = None,
               post_generation_check_function: Callable = None,
               post_evolution_function: Callable = None,
               ) -> Individual:
        """
        Starts the evolutionary run.

        Args:
            starting_generation (int): Starting generation (default = 0).
            n_generations (int): Number of generations to run (default = 5).
            crossover_probability (float): Initial crossover probability (default = 0.5).
            mutation_probability (float): Initial mutation probability (default = 0.25).
            crossover_params (dict): Dict of params for custom crossover function (default = None).
            mutation_params (dict): Dict of params for custom mutation function (default = None).
            parent_selection_params (dict): Dict of params for custom selection function (default = None).
            parent_combination_params (dict): Dict of params for custom combination function (default = None).
            survivor_selection_params (dict): Dict of params for custom survivor selection function (default = None).
            population_growth_params (dict): Dict of params for custom population growth function (defualt = None).
            alien_spawn_params (dict): Dict of params for alien spawn function (default = None).
            criterion_function (Callable): A function to evaluate if the desired criterion has been met (default = None).
            criterion_params (dict): Function parameters for criterion (default = None).
            pre_generation_check_function (Callable): A function to perform some custom pre-action at the start of every generation, with signature ``func(generation, island)`` (default = None).
            post_generation_check_function (Callable): A function to perform some custom post-action at the end of every generation, with signature ``func(generation, island)``  (default = None).
            post_evolution_function (Callable): A function to perform some custom post-action after full evolution cycle, with signature ``func(island)`` (default = None).

        Returns:
            Individual: The fittest Individual found.
        """

        if crossover_params:
            _crossover_params = crossover_params
        else:
            _crossover_params = {}

        if mutation_params:
            _mutation_params = mutation_params
        else:
            _mutation_params = {}

        if parent_selection_params:
            _parent_selection_params = parent_selection_params
        else:
            _parent_selection_params = {}

        if parent_combination_params:
            _parent_combination_params = parent_combination_params
        else:
            _parent_combination_params = {}

        if survivor_selection_params:
            _survivor_selection_params = survivor_selection_params
        else:
            _survivor_selection_params = {}

        if population_growth_params:
            _population_growth_params = population_growth_params
        else:
            _population_growth_params = {}

        if alien_spawn_params:
            _alien_spawn_params = alien_spawn_params
        else:
            _alien_spawn_params = {}

        def _default_criterion_function(island):
            return island.generation_count < starting_generation + n_generations - 1

        if criterion_function:
            g_func = criterion_function
        else:
            g_func = _default_criterion_function
            criterion_params = {}

        g = starting_generation
        self.verbose_logging(f"evolve: start_generation {g}")
        while g_func(island=self, **criterion_params):
            self.verbose_logging(f"evolve: generation_number {g}")

            if pre_generation_check_function:
                pre_generation_check_function(generation=g, island=self)

            self.__evolutionary_engine(g=g,
                                       parent_selection_params=_parent_selection_params,
                                       parent_combination_params=_parent_combination_params,
                                       survivor_selection_params=_survivor_selection_params,
                                       population_growth_params=_population_growth_params,
                                       crossover_probability=crossover_probability,
                                       mutation_probability=mutation_probability,
                                       crossover_params=_crossover_params,
                                       mutation_params=_mutation_params,
                                       alien_spawn_params=_alien_spawn_params)
            if post_generation_check_function:
                post_generation_check_function(generation=g, island=self)
            g += 1


        best_ind = selection_elites_top_n(island=self, individuals=self.population, n=1)[0]

        self.verbose_logging(f"evolve: end")
        self.verbose_logging(f"evolve: best {str(best_ind)}")

        if post_evolution_function:
            post_evolution_function(island=self)

        return best_ind

    def save_island(self, filepath : str):
        """
        Dumps a pickled self to the given file path.

        Args:
            filepath (str): File path to pickled island.
        """
        with open(filepath, "wb") as f:
            pickle.dump(self.__dict__, f)
        self.verbose_logging(f"save: file {filepath}")

    def load_island(self, filepath : str):
        """
        Loads a pickled self from the given file path.

        Args:
            filepath (str): File path to pickled island.
        """
        with open(filepath, "rb") as f:
            self.__dict__.update(pickle.load(f))
        self.verbose_logging(f"load: file {filepath}")

    def __evaluate_individuals(self, individuals):
        if self.core_count == 1:
            return self.population_evaluation_function(individuals=individuals, island=self, params=self.function_params)

        with mp.Pool(self.core_count) as p:
            for f, i in zip(p.starmap(evaluate_individual_multiproc_wrapper, [(i, self, self.function_params) for i in individuals]), individuals):
                i.fitness = f

        return individuals

    def __evolutionary_engine(self,
                              g,
                              parent_selection_params,
                              parent_combination_params,
                              survivor_selection_params,
                              population_growth_params,
                              crossover_probability,
                              mutation_probability,
                              crossover_params,
                              mutation_params,
                              alien_spawn_params):

        if self.save_checkpoint_level == 1:
            self.save_checkpoint(event=f'evolve_pre_{g}', island=self)

        selected_parents = self.parent_selection(island=self,
                                       individuals=self.clone(individuals=self.population, island=self),
                                       **parent_selection_params)
        self.verbose_logging(f"select: selected_parents_count {len(selected_parents)}")

        self.parents.append({'generation' : g, 'parents' : selected_parents})

        # Children are strictly copies or new objects seeing as they have a lineage and parents
        generation_children = list()
        for parents in self.parent_combination(individuals=selected_parents, island=self, **parent_combination_params):
            self.verbose_logging(f"select: parent_count {len(parents)}")
            if self.crossover_prob(crossover_probability=crossover_probability, island=self):
                self.verbose_logging(f"cross: parents {[str(p) for p in parents]}")

                children = self.crossover(island=self,
                                          individuals=self.clone(individuals=parents, island=self),
                                          **crossover_params)


                for child in children:
                    child.reset_name()
                    child.register_parent_names(parents)
                    child.age = 0

                self.verbose_logging(f"cross: offspring_count {len(children)}")
                self.verbose_logging(f"cross: offspring {[str(p) for p in children]}")

                generation_children.extend(children)

        self.children.append({'generation' : g, 'children' : generation_children})

        # Mutants are not strictly copied but rather only modified seeing as they are part of the children list
        generation_mutants = list()
        for mutant in generation_children:
            if self.mutation_prob(mutation_probability=mutation_probability, island=self):
                self.verbose_logging(f"mutate: offspring {str(mutant)}")

                mutated = self.mutation(island=self, individual=mutant, **mutation_params)

                generation_mutants.append(mutated)

        self.mutants.append({'generation': g, 'mutants': generation_mutants})

        offspring_fitnesses = list()

        if self.save_checkpoint_level == 2:
            self.save_checkpoint(event=f'evolve_pre_eval_{g}', island=self)

        generation_children = self.__evaluate_individuals(generation_children)

        for individual in generation_children:
            offspring_fitnesses.append(individual.fitness)

        if self.save_checkpoint_level == 2:
            self.save_checkpoint(event=f'evolve_post_eval_{g}', island=self)

        for child in generation_children:
            self.__add_to_lineage(child)
            for p in child.parents:
                edge = {
                    'from': p.name,
                    'to': child.name,
                    'generation' : g
                }
                gene_inheritance = child.get('gene_inheritance')
                if not gene_inheritance is None:
                    edge['gene_inheritance'] = gene_inheritance[p.name]
                self.lineage['edges'].append(edge)

        interim_children = list()
        for individual in self.survivor_selection(individuals=generation_children, island=self, **survivor_selection_params):
            if self.allow_twins or not individual.unique_genetic_code(force_update=True) in self.unique_genome:
                # If we want a diverse gene pool, this must be true
                interim_children.append(individual)
                self.verbose_logging(f"evolve: add {str(individual)}")
                self.unique_genome.append(individual.unique_genetic_code())

        grown_population, deaths = self.population_growth_function(population=self.population,
                                                                  offspring=interim_children,
                                                                  island=self, **population_growth_params)
        if len(grown_population) == 0:
            raise IslandError(f"Population length of 0. {len(interim_children)} offspring were created/survived in generation {g}")
        self.population = grown_population
        self.deaths[g] = deaths

        population_fitnesses = [ind.fitness for ind in self.population]
        selected_parents_fitnesses = [ind.fitness for ind in selected_parents]

        if len(offspring_fitnesses) > 0:
            self.generation_info.append(
                {
                    self.__generation_key: g,
                    self.__stat_key: 'offspring',
                    self.__pop_len_key: len(offspring_fitnesses),
                    self.__fitness_mean_key: np.mean(offspring_fitnesses),
                    self.__fitness_std_key: np.std(offspring_fitnesses),
                    self.__fitness_min_key: min(offspring_fitnesses),
                    self.__fitness_max_key: max(offspring_fitnesses),
                    self.__most_fit_key: selection_elites_top_n(island=self, individuals=generation_children, n=1)[0].name,
                    self.__least_fit_key : selection_elites_top_n(island=self, individuals=generation_children, n=1)[-1].name
                }
            )
            self.verbose_logging(f"evolve: stats {self.generation_info[-1]}")

        self.generation_info.append(
            {
                self.__generation_key: g,
                self.__stat_key: 'parents',
                self.__pop_len_key: len(selected_parents_fitnesses),
                self.__fitness_mean_key: np.mean(selected_parents_fitnesses),
                self.__fitness_std_key: np.std(selected_parents_fitnesses),
                self.__fitness_min_key: min(selected_parents_fitnesses),
                self.__fitness_max_key: max(selected_parents_fitnesses),
                self.__most_fit_key: selection_elites_top_n(island=self, individuals=selected_parents, n=1)[0].name,
                self.__least_fit_key : selection_elites_top_n(island=self, individuals=selected_parents, n=1)[-1].name
            }
        )


        self.verbose_logging(f"evolve: stats {self.generation_info[-1]}")

        pop_stats = selection_elites_top_n(island=self,
                                           individuals=copy.deepcopy(self.population),
                                           n=len(self.population))

        self.generation_info.append(
            {
                self.__generation_key: g,
                self.__stat_key: 'population',
                self.__pop_len_key: len(population_fitnesses),
                self.__fitness_mean_key: np.mean(population_fitnesses),
                self.__fitness_std_key: np.std(population_fitnesses),
                self.__fitness_min_key: min(population_fitnesses),
                self.__fitness_max_key: max(population_fitnesses),
                self.__most_fit_key: pop_stats[0].name,
                self.__least_fit_key: pop_stats[-1].name
            }
        )

        self.verbose_logging(f"evolve: stats {self.generation_info[-1]}")

        elites = pop_stats[:4]
        elite_fitnesses = [ind.fitness for ind in elites]
        self.elites.append({'generation': g, 'elites': elites})

        self.generation_info.append(
            {
                self.__generation_key: g,
                self.__stat_key: 'elites',
                self.__pop_len_key: 4,
                self.__fitness_mean_key: np.mean(elite_fitnesses),
                self.__fitness_std_key: np.std(elite_fitnesses),
                self.__fitness_min_key: min(elite_fitnesses),
                self.__fitness_max_key: max(elite_fitnesses),
                self.__most_fit_key: elites[0].name,
                self.__least_fit_key: elites[-1].name
            }
        )

        self.verbose_logging(f"evolve: stats {self.generation_info[-1]}")

        for i in self.population:
            i.birthday()

        new_aliens = self.alien_spawn_function(**alien_spawn_params, island=self)

        if len(new_aliens) > 0:
            alien_fitnesses = list()
            for alien in self.__evaluate_individuals(new_aliens):
                if self.allow_twins or not alien.unique_genetic_code(force_update=True) in self.unique_genome:
                    # Else, add it effectively allowing "twins" to exist
                    alien_fitnesses.append(alien.fitness)
                    self.verbose_logging(f"evolve: add {str(alien)}")
                    alien.history.append({"alien_init_island": self.name})
                    self.population.append(alien)
                    self.unique_genome.append(alien.unique_genetic_code())
                    self.__add_to_lineage(alien)

            self.generation_info.append(
                {
                    self.__generation_key: g,
                    self.__stat_key: 'aliens',
                    self.__pop_len_key: len(alien_fitnesses),
                    self.__fitness_mean_key: np.mean(alien_fitnesses),
                    self.__fitness_std_key: np.std(alien_fitnesses),
                    self.__fitness_min_key: min(alien_fitnesses),
                    self.__fitness_max_key: max(alien_fitnesses),
                    self.__most_fit_key: selection_elites_top_n(island=self, individuals=new_aliens, n=1)[0].name,
                    self.__least_fit_key: selection_elites_top_n(island=self, individuals=new_aliens, n=1)[-1].name
                }
            )

        self.generation_count = g

        if self.save_checkpoint_level == 1:
            self.save_checkpoint(event=f'evolve_post_{g}', island=self)

    def write_lineage_json(self, filename : str):
        """
        Dumps the lineage safely to JSON file.

        Args:
            filename (str): Output file.
        """
        nodes = list()
        for n in self.lineage['nodes']:
            nodes.append({k:v for k, v in n.items() if not k.startswith('_')})
        import json
        with open(filename, 'w', newline='', encoding='utf8') as output_file:
            json.dump({'nodes' : nodes, 'edges' : self.lineage['edges']}, output_file)


    def write_report(self, filename : str, output_json : bool = False):
        """
        Write the generational history to CSV (or JSON) file.

        Args:
            filename (str): Output file.
            output_json (bool): Write as JSON instead of CSV (default = False).
        """
        if output_json:
            import json
            with open(filename, 'w', newline='', encoding='utf8') as output_file:
                json.dump(self.generation_info, output_file)
        else:
            import csv
            keys = [
                self.__generation_key,
                self.__stat_key,
                self.__pop_len_key,
                self.__fitness_mean_key,
                self.__fitness_std_key,
                self.__fitness_min_key,
                self.__fitness_max_key,
                self.__most_fit_key,
                self.__least_fit_key
            ]
            with open(filename, 'w', newline='', encoding='utf8') as output_file:
                dict_writer = csv.DictWriter(output_file, keys)
                dict_writer.writeheader()
                dict_writer.writerows(self.generation_info)

    def add_new_property(self, key : str, value : Any):
        """
        Method to add new properties (attributes).

        Args:
            key (str): Name of property.
            value (Any): Anything.
        """
        if self.__island_properties:
            self.__island_properties.update({key: value})
        else:
            self.__island_properties = {key: value}
        self.__dict__.update({key: value})

    def get_properties(self):
        """
        Gets a dict of the custom properties that were added at initialisation or the `add_new_property` method.

        Returns:
            dict: All custom properties.
        """
        return self.__island_properties

    def get(self, key: str, default=None):
        """
        Gets the value of a property or returns default if it doesn't exist.

        Args:
            key (str): Property name.
            default: Value to return if the property is not found (default = None).

        Returns:
            any: The property of the individual.
        """
        if key in self.__dict__.keys():
            return self.__dict__[key]
        return default

    def verbose_logging(self, event_message):
        if self.verbose:
            logging.info(event_message, extra={'island' : self.name})
        if self.logging_function:
            self.logging_function(event_message=event_message, island=self)




import unittest
import natural_selection.genetic_programs.node_operators as ops
from natural_selection.genetic_programs import GeneticProgram
from natural_selection.genetic_programs.operators.initialisation import initialise_population_grow_method
from natural_selection.genetic_programs.operators.crossover import crossover_two_one_point
from natural_selection.genetic_programs.operators.mutation import mutation_random_point_change
from natural_selection import Island

import numpy as np


def fitness_function(program, island, X, Y):
    try:
        Y_pred = [program(x=x) for x in X]

        f = np.square(np.subtract(Y, Y_pred)).mean()

        f = f if f > 0 else 999999
    except Exception as ex:
        print(ex)
        f = 999999

    return f

class TestGeneticProgram(unittest.TestCase):

    def test_island_symbolic_regression(self):

        # Let's try find a simple quadratic

        x = list(range(5))
        ground_truth = list(map(lambda x: x**2 + x, x))





        operators = [ops.OperatorPow, ops.OperatorAdd]
        terminals = ['x', 2]

        gp_symb = GeneticProgram(fitness_function=fitness_function, operators=operators, terminals=terminals, max_depth=3, min_depth=2, growth_mode='grow')


        island = Island(function_params={'X':x,'Y':ground_truth},
                        maximise_function=False,
                        initialisation_function=initialise_population_grow_method,
                        crossover_function=crossover_two_one_point,
                        mutation_function=mutation_random_point_change,
                        core_count=1,
                        verbose=False)
        island.initialise(gp_symb, population_size=4)

        best = island.evolve(mutation_probability=0.1, n_generations=50)

        self.assertIsInstance(best, GeneticProgram)

        Y_pred = [best(x=v) for v in x]

        self.assertEqual(len(Y_pred), len(ground_truth))






import copy
import warnings as w

def population_steady_state_remove_weakest(population, offspring, island, desc=True):
    """
    A steady-state method where offspring members replace the weakest population members.

    Args:
        population (list): The current population to add to or replace.
        offspring (list): The offspring members to add.
        island (Island): The calling Island.
        desc (bool): If used outside Island (default = True).

    Returns:
        tuple: The new population, the removed members
    """
    if len(offspring) == 0:
        return population, list()
    cloned = copy.deepcopy(population)
    def sortFitness(val):
        return val.fitness

    if island:
        cloned.sort(key=sortFitness, reverse=island.maximise_function)
    else:
        cloned.sort(key=sortFitness, reverse=desc)

    deaths = list(cloned[-len(offspring):])
    cloned[-len(offspring):] = offspring
    return cloned, deaths

def population_steady_state_remove_oldest(population, offspring, island):
    """
    A steady-state method where offspring members replace the oldest members.

    Args:
        population (list): The current population to add to or replace.
        offspring (list): The offspring members to add.
        island (Island): The calling Island.

    Returns:
        tuple: The new population, the removed members
    """
    if len(offspring) == 0:
        return population, list()
    cloned = copy.deepcopy(population)
    def sortAge(val):
        return val.age

    cloned.sort(key=sortAge, reverse=False)

    deaths = list(cloned[-len(offspring):])
    cloned[-len(offspring):] = offspring
    return cloned, deaths

def population_generational(population, offspring, island):
    """
    The generational method where the offspring replaces the full population.

    Args:
        population (list): The current population to add to or replace.
        offspring (list): The offspring members to add.
        island (Island): The calling Island.

    Returns:
        tuple: The new population, the removed members
    """
    if len(population) != len(offspring):
        w.warn(f"WARNING: offspring size {len(offspring)} != population size {len(population)}. Set selection count to population size!")
    return offspring, copy.deepcopy(population)

def population_incremental(population, offspring, island):
    """
    A simple incremental run where offspring members are added to the population.

    Args:
        population (list): The current population to add to or replace.
        offspring (list): The offspring members to add.
        island (Island): The calling Island.

    Returns:
        tuple: The new population, the removed members
    """
    new_population = copy.deepcopy(population)
    new_population.extend(offspring)
    return new_population, list()



import copy
import os
from datetime import datetime

def clone_classic(individuals : list, island=None):
    """
    Classic cloning function, making a deep copy of population list.

    Args:
        individuals (list): Population members to copy.
        island (Island): The Island calling the method (optional, default = None).

    Returns:
        list: Carbon copy of population.
    """
    return copy.deepcopy(individuals)

def get_random_string(length : int = 8, include_numeric=False) -> str:
    """
    Generate a random string with a given length. Used mainly for password generation.
    Args:
        length (int): Length to generate.
        include_numeric (bool): Include numbers?

    Returns:
        str: Random character string.
    """
    import random
    import string

    letters = string.ascii_letters
    if include_numeric:
        letters = f'{letters}{string.digits}'
    return ''.join(random.choice(letters) for i in range(length))

def default_save_checkpoint_function(event, island):
    """
    Default checkpoint saving function. Used to save the Island periodically after certain events.

    Args:
        event (str): The evolutionary event (crossover, eval etc).
        island (Island): The island to save.

    """
    if not os.path.isdir(island.checkpoints_dir):
        os.mkdir(island.checkpoints_dir)
    if not os.path.isdir(f'{island.checkpoints_dir}/{island.name}'):
        os.mkdir(f'{island.checkpoints_dir}/{island.name}')
    fp = f'{island.checkpoints_dir}/{island.name}/{datetime.utcnow().strftime("%H-%M-%S")}_{event}_checkpoint.pkl'
    island.verbose_logging(f'checkpoint: file {fp}')
    island.save_island(fp)

def post_evolution_function_save_all(island):
    """
    Simple function to save (pickle) every population member after completing evolution.

    Args:
        island (Island): Island to get ``population`` from.
    """
    for p in island.population:
        p.save_individual(filepath=f'individual_{p.name}.pkl')

def evaluate_individual_multiproc_wrapper(individual, island, params):
    """
    A wrapper function for spawning multi processes.

    Args:
        individual: Single individual to evaluate.
        island: The calling island
        params: Function eval params.

    Returns:
        numeric : the fitness value.
    """
    island.verbose_logging(f"eval: {str(individual)}")
    return individual.evaluate(params=params, island=island)

def evaluate_individuals_sequentially(individuals, island, params):
    """
    A simple function that wraps the sequential evaluation of population members, all locally.

    Args:
        individuals (list): The individuals to evaluate.
        island (Island): The calling island.
        params (dict): The evaluation parameters.

    Returns:

    """
    for individual in individuals:
        island.verbose_logging(f"eval: {str(individual)}")
        individual.evaluate(island=island, params=params)
    return individuals

class IslandError(Exception):
    """
    Encapsulating graceful exception handling during evolutionary runs.

    Args:
        message (str): Message to print.
        exit (bool): Whether to hard exit the process or not (default = False).
    """

    def __init__(self, message : str, exit : bool = False):
        self.message = message
        if exit:
            print(f"IslandError: {self.message}")
            quit(1)


    def __str__(self):
        return f"IslandError: {self.message}"


import unittest
from natural_selection.genetic_programs.node_operators import OperatorAdd, OperatorSub, OperatorMul, OperatorDiv, Operator
import natural_selection.genetic_programs.node_operators as op
from natural_selection.genetic_programs import Node, GeneticProgram

class TestNodeOperators(unittest.TestCase):

    def test_custom_operator_with_function(self):
        def func(args):
            return abs(args[0] - args[1]) / (args[0] + args[1])

        custom = Operator(operator_label='custom',function=func,min_arity=2,max_arity=2)


        returned_value = custom.exec([15,20])
        expected_value = abs(15 - 20 ) / (15 + 20)

        self.assertEqual(returned_value,expected_value)

        with self.assertRaises(AssertionError):
            returned_value = custom.exec([15,20,50])

        with self.assertRaises(AssertionError):
            returned_value = custom.exec([15])

        n1 = Node(label='X', is_terminal=True)
        n2 = Node(label='Y', is_terminal=True)

        n = Node(arity=2,operator=custom, children=[n1, n2])

        X = 16
        Y = 19

        expected_value = abs(X - Y) / (X + Y)
        returned_value = n(X=X,Y=Y)

        self.assertEqual(returned_value, expected_value)




# -*- coding: utf-8 -*-
"""Basic classes for running Genetic Algorithms.
"""
__author__ = "Justin Hocking"
__copyright__ = "Copyright 2022, Zipfian Science"
__credits__ = []
__license__ = ""
__version__ = "0.2.0"
__maintainer__ = "Justin Hocking"
__email__ = "justin.hocking@zipfian.science"
__status__ = "Development"


import uuid
from typing import Callable, Any, Iterable
import pickle
from collections import OrderedDict
import copy
import warnings as w

from natural_selection.genetic_algorithms.utils import GeneticAlgorithmError


class Gene:
    """
    A simple class to encapsulate a simple gene.

    Args:
        name (str): Gene name. The gene name also acts as a compatibility reference.
        value (Any): The value, could be any type.
        randomise_function (Callable): A function to randomise the gene, taking the gene (`self`) as input with signature ``func(self)``.
        gene_max (Any, numeric type): Max value for random number generator (default = None).
        gene_min (Any, numeric type): Min value for random number generator (default = None).
        mu (Any, numeric type): Mean value of distribution to sample from (default = 0).
        sig (Any, numeric type): Std. Dev. value of distribution to sample from (default = 1).
        step_lower_bound (Any, numeric type): For uniform stepping functions, defines lower bound of range (default = -1.0).
        step_upper_bound (Any, numeric type): For uniform stepping functions, defines upper bound of range (default = 1.0).
        choices (Iterable): List of choices, categorical or not, to randomly choose from (default = None).
        gene_properties (dict): For custom random functions, extra params may be given (default = None).

    """

    def __init__(self,
                 name : str,
                 value : Any,
                 randomise_function: Callable,
                 gene_max : Any = None,
                 gene_min : Any = None,
                 mu : Any = 0,
                 sig: Any = 1,
                 step_lower_bound : Any = -1.0,
                 step_upper_bound: Any = 1.0,
                 choices : Iterable  = None,
                 gene_properties : dict = None):
        if '<lambda>' in repr(randomise_function):
            w.warn("WARNING: 'randomise_function' lambda can not be pickled using standard libraries.")
        self.name = name
        self.value = value
        self.gene_max = gene_max
        self.gene_min = gene_min
        self.mu = mu
        self.sig = sig
        self.step_lower_bound = step_lower_bound
        self.step_upper_bound = step_upper_bound
        self.choices = choices
        self.randomise_function = randomise_function
        self.__gene_properties = gene_properties
        if gene_properties:
            for k, v in gene_properties.items():
                self.__dict__.update({k:v})

    def randomise(self):
        """
        Sets a random value gene with randomised value.
        """
        self.value = self.randomise_function(gene=self)

    def randomise_new(self):
        """
        Return a new gene with randomised value.

        Returns:
            Gene: Newly created gene.
        """
        return Gene(
            name=copy.copy(self.name),
            value=self.randomise_function(gene=self),
            randomise_function=self.randomise_function,
            gene_max=copy.copy(self.gene_max),
            gene_min=copy.copy(self.gene_min),
            mu=copy.copy(self.mu),
            sig=copy.copy(self.sig),
            choices=copy.deepcopy(self.choices),
            gene_properties=copy.deepcopy(self.__gene_properties)
        )

    def add_new_property(self, key : str, value : Any):
        """
        Method to add new properties (attributes).

        Args:
            key (str): Name of property.
            value (Any): Anything.
        """
        if self.__gene_properties:
            self.__gene_properties.update({key: value})
        else:
            self.__gene_properties = {key: value}
        self.__dict__.update({key: value})

    def get_properties(self):
        """
        Gets a dict of the custom properties that were added at initialisation or the `add_new_property` method.

        Returns:
            dict: All custom properties.
        """
        return self.__gene_properties

    def get(self, key: str, default=None):
        """
        Gets the value of a property or returns default if it doesn't exist.

        Args:
            key (str): Property name.
            default: Value to return if the property is not found (default = None).

        Returns:
            any: The property of the individual.
        """
        if key in self.__dict__.keys():
            return self.__dict__[key]
        return default

    def __str__(self) -> str:
        return f'Gene({self.name}:{self.value})'

    def __repr__(self):
        start_str = f'Gene({self.name}:{self.value}:{self.gene_max}:{self.gene_min}:{self.mu}:{self.sig}:{self.step_lower_bound}:{self.step_upper_bound}'
        if self.choices:
            start_str = f"{start_str}:[{repr(self.choices)}]"
        start_str = f'{start_str}:{self.randomise_function.__name__}'
        if self.__gene_properties:
            start_str = f'{start_str}:{str(self.__gene_properties)}'
        return  f'{start_str})'


class Chromosome:
    """
    A class that encapsulates an ordered sequence of Gene objects.

    Note:
        gene_verify_func should take the gene, index of gene, and chromosome (`self`) as parameters.

    Args:
        genes (list): list of initialised Gene objects.
        gene_verify_func (Callable): A function to verify gene compatibility `func(gene,loc,chromosome)` (default = None).
        chromosome_properties (dict): For custom functions, extra params may be given (default = None).
    """

    def __init__(self, genes: list = None,
                 gene_verify_func : Callable = None,
                 chromosome_properties : dict = None):
        if genes:
            self.genes = genes
        else:
            self.genes = list()

        if gene_verify_func and '<lambda>' in repr(gene_verify_func):
            w.warn("WARNING: 'gene_verify_func' lambda can not be pickled using standard libraries.")
        self.gene_verify_func = gene_verify_func
        self.__chromosome_properties = chromosome_properties
        if chromosome_properties:
            for k, v in chromosome_properties.items():
                self.__dict__.update({k: v})

    def append(self, gene: Gene):
        """
        Simple appending of Gene type objects.

        Args:
            gene (Gene): Gene
        """
        assert isinstance(gene, Gene), 'Must be Gene type!'
        if self.gene_verify_func and not self.gene_verify_func(gene=gene,loc=-1,chromosome=self):
            raise GeneticAlgorithmError(message="Added gene did not pass compatibility tests!")
        self.genes.append(gene)

    def add_new_property(self, key : str, value : Any):
        """
        Method to add new properties (attributes).

        Args:
            key (str): Name of property.
            value (Any): Anything.
        """
        if self.__chromosome_properties:
            self.__chromosome_properties.update({key: value})
        else:
            self.__chromosome_properties = {key: value}
        self.__dict__.update({key: value})

    def get_properties(self):
        """
        Gets a dict of the custom properties that were added at initialisation or the `add_new_property` method.

        Returns:
            dict: All custom properties.
        """
        return self.__chromosome_properties

    def get(self, key : str, default=None):
        """
        Gets the value of a property or returns default if it doesn't exist.

        Args:
            key (str): Property name.
            default: Value to return if the property is not found (default = None).

        Returns:
            any: The property of the individual.
        """
        if key in self.__dict__.keys():
            return self.__dict__[key]
        return default

    def randomise_gene(self, index : int):
        """
        Randomise a gene at index.

        Args:
            index (int): Index of gene.
        """
        assert index < len(self.genes), 'Index Out of bounds!'
        self.genes[index].randomise()

    def randomise_all_genes(self):
        """
        Randomises all genes in chromosome.
        """
        for gene in self.genes:
            gene.randomise()

    def __setitem__(self, index, gene):
        if isinstance(index, slice):
            assert index.start < len(self.genes), 'Index Out of bounds!'
        else:
            assert isinstance(gene, Gene), 'Must be Gene type!'
            assert index < len(self.genes), 'Index Out of bounds!'

        if self.gene_verify_func and not self.gene_verify_func(gene=gene, loc=index, chromosome=self):
            raise GeneticAlgorithmError("Index set gene did not pass compatibility tests!")
        self.genes[index] = gene

    def to_dict(self) -> OrderedDict:
        """
        Helper function to convert chromosome into a key-value Python dictionary, assuming genes have unique names!

        Returns:
            OrderedDict: Ordered dictionary of genes.
        """
        return OrderedDict({gene.name : gene.value for gene in self.genes})

    def to_list(self) -> list:
        """
        Helper function to convert chromosome into a Python list of the gene values.

        Returns:
            list: Values of the genes in order, as a list.
        """
        return [gene.value for gene in self.genes]


    def __getitem__(self, index) -> Gene:
        if isinstance(index, slice):
            assert index.start < len(self.genes), 'Index Out of bounds!'
        else:
            assert index < len(self.genes), 'Index Out of bounds!'

        return self.genes[index]

    def __iter__(self):
        self.__n = 0
        return self

    def __next__(self):
        if self.__n < len(self.genes):
            gene = self.genes[self.__n]
            self.__n += 1
            return gene
        else:
            raise StopIteration

    def __len__(self):
        return len(self.genes)

    def __str__(self) -> str:
        start_str = '-'.join([str(t) for t in self.genes])
        return f'Chromosome({start_str})'

    def __repr__(self):
        start_str = '-'.join([repr(t) for t in self.genes])
        start_str = f'Chromosome({start_str})'
        return start_str


class Individual:
    """
    A class that encapsulates a single individual, with genetic code and a fitness evaluation function.

    Args:
        fitness_function (Callable): Function with ``func(Chromosome, island, **params)`` signature (default = None).
        name (str): Name for keeping track of lineage (default = None).
        chromosome (Chromosome): A Chromosome object, initialised (default = None).
        species_type (str) : A unique string to identify the species type, for preventing cross polluting (default = None).
        filepath (str): Skip init and load from a pickled file.
        individual_properties (dict): For fitness functions, extra params may be given (default = None).

    Attributes:
        fitness (Numeric): The fitness score after evaluation.
        age (int): How many generations was the individual alive.
        genetic_code (str): String representation of Chromosome.
        history (list): List of dicts of every evaluation.
        parents (list): List of strings of parent names.
    """

    def __init__(self, fitness_function : Callable = None,
                 name : str = None,
                 chromosome: Chromosome = None,
                 species_type : str = None,
                 filepath : str = None,
                 individual_properties : dict = None):
        if not filepath is None:
            self.load(filepath=filepath)
            return
        if fitness_function and '<lambda>' in repr(fitness_function):
            w.warn("WARNING: 'fitness_function' lambda can not be pickled using standard libraries.")
        if name is None:
            self.name = str(uuid.uuid4())
        else:
            self.name = name
        if chromosome is None:
            self.chromosome = Chromosome()
        else:
            self.chromosome = chromosome
        self.fitness_function = fitness_function
        self.fitness = None
        self.age = 0
        self.genetic_code = None
        self.history = list()
        self.parents = list()
        if species_type:
            self.species_type = species_type
        else:
            self.species_type = "def"

        self.__individual_properties = individual_properties
        if individual_properties:
            for k, v in individual_properties.items():
                self.__dict__.update({k: v})

    def register_parent_names(self, parents : list, reset_parent_name_list : bool = True):
        """
        In keeping lineage of family lines, the names of parents are kept track of.

        Args:
            parents (list): A list of Individuals of the parents.
        """
        if reset_parent_name_list:
            self.parents = list()
        for parent in parents:
            self.parents.append(parent)

    def reset_name(self, name : str = None):
        """
        A function to reset the name of an individual, helping to keep linage of families.

        Args:
            name (str): Name (default = None).
        """
        if name is None:
            self.name = str(uuid.uuid4())
        else:
            self.name = name

    def birthday(self, add : int = 1):
        """
        Add to the age. This is for keeping track of how many generations an individual has "lived" through.

        Args:
            add (int): Amount to age.
        """
        self.age += add

    def reset_fitness(self, fitness : Any = None, reset_genetic_code : bool = True):
        """
        Reset (or set) the fitness of the individual.

        Args:
            fitness (Any): New fitness value (default = None).
            reset_genetic_code (bool): Whether to reset the genetic code. (default = True)
        """
        self.fitness = fitness
        if reset_genetic_code:
            self.genetic_code = None

    def add_gene(self, gene : Gene):
        """
        Appends a gene to the chromosome.

        Args:
            gene (Gene): Gene to add.
        """
        self.chromosome.append(gene)

    def evaluate(self, params : dict = None, island=None) -> Any:
        """
        Run the fitness function with the given params.

        Args:
            params (dict): Named dict of eval params (default = None).
            island (Island): Pass the Island for advanced fitness functions based on Island properties and populations (default = None).

        Returns:
            numeric: Fitness value.
        """
        if not params is None:
            _params = params
        else:
            _params = {}
        try:
            self.fitness = self.fitness_function(individual=self, island=island, **_params)
        except Exception as exc:
            if island:
                island.verbose_logging(f"ERROR: {self.name} - {repr(self.chromosome)} - {repr(exc)}")
            raise GeneticAlgorithmError(message=f'Could not evaluate individual "{self.name}" due to {repr(exc)}')

        stamp = { "name": self.name,
                  "age": self.age,
                  "fitness": self.fitness,
                  "chromosome": str(self.chromosome),
                  "parents": self.parents,
         }

        if island:
            stamp["island_generation" ] = island.generation_count

        self.history.append(stamp)
        return self.fitness

    def unique_genetic_code(self, force_update : bool = False) -> str:
        """
        Gets the unique genetic code, generating if it is undefined.

        Args:
            force_update (bool): Force update of genetic_code property (default = False).

        Returns:
            str: String name of Chromosome.
        """
        if self.genetic_code is None or force_update:
            self.genetic_code = repr(self.chromosome)
        return self.genetic_code

    def save(self, filepath : str):
        """
        Save an individual to a pickle file.

        Args:
            filepath (str): File path to write to.
        """
        with open(filepath, "wb") as f:
            pickle.dump(self.__dict__, f)

    def load(self, filepath : str):
        """
        Load an individual from a pickle file.

        Args:
            filepath (str): File path to load from.
        """
        with open(filepath, "rb") as f:
            self.__dict__.update(pickle.load(f))

    def add_new_property(self, key : str, value : Any):
        """
        Method to add new properties (attributes).

        Args:
            key (str): Name of property.
            value (Any): Anything.
        """
        if self.__individual_properties:
            self.__individual_properties.update({key: value})
        else:
            self.__individual_properties = {key: value}
        self.__dict__.update({key: value})

    def get_properties(self) -> dict:
        """
        Gets a dict of the custom properties that were added at initialisation or the `add_new_property` method.

        Returns:
            dict: All custom properties.
        """
        return self.__individual_properties

    def get(self, key : str, default=None):
        """
        Gets the value of a property or returns default if it doesn't exist.

        Args:
            key (str): Property name.
            default: Value to return if the property is not found (default = None).

        Returns:
            any: The property of the individual.
        """
        if key in self.__dict__.keys():
            return self.__dict__[key]
        return default

    def __str__(self) -> str:
        return f'Individual({self.name}:{self.fitness})'

    def __repr__(self) -> str:
        genetic_code = self.unique_genetic_code()
        return f'Individual({self.name}:{self.fitness}:{self.age}:{self.species_type}:{genetic_code})'

    def __eq__(self, other):
        if isinstance(other, Individual):
            return self.unique_genetic_code() == other.unique_genetic_code()
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __ne__(self, other):
        if isinstance(other, Individual):
            return self.unique_genetic_code() != other.unique_genetic_code()
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __lt__(self, other):
        if isinstance(other, Individual):
            return self.fitness < other.fitness
        elif isinstance(other, int):
            return self.fitness < other
        elif isinstance(other, float):
            return self.fitness < other
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __le__(self, other):
        if isinstance(other, Individual):
            return self.fitness <= other.fitness
        elif isinstance(other, int):
            return self.fitness <= other
        elif isinstance(other, float):
            return self.fitness <= other
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __gt__(self, other):
        if isinstance(other, Individual):
            return self.fitness > other.fitness
        elif isinstance(other, int):
            return self.fitness > other
        elif isinstance(other, float):
            return self.fitness > other
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')

    def __ge__(self, other):
        if isinstance(other, Individual):
            return self.fitness >= other.fitness
        elif isinstance(other, int):
            return self.fitness >= other
        elif isinstance(other, float):
            return self.fitness >= other
        else:
            raise GeneticAlgorithmError(message=f'Can not compare {type(other)}')




import random

def mutation_random_point_change(individual, prob : float = None, island=None):
    """
    A Classic mutation function, changes a random subtree of the given individual.

    Args:
        individual (GeneticProgram): Individual object containing a Node tree.
        prob (float): Not used in GP mutation (default = None).
        island (Island): The Island calling the method (default = None).

    Returns:
        GeneticProgram: The newly mutated program.
    """

    tree = individual.node_tree


    tree_depth = tree.depth()

    depth = random.randint(2, tree_depth - 1) if tree_depth > 2 else 2

    breadth = tree.breadth(depth=depth)

    random_point = random.randint(0, breadth - 1)

    max_depth = individual.max_depth - depth + 1

    new_subtree = individual.tree_generator(operators=individual.operators,
            terminals=individual.terminals,
            max_depth=max_depth,
            min_depth=1,
            growth_mode=individual.growth_mode,
            terminal_prob=individual.terminal_prob,
            genetic_program=individual
        )

    tree.set_subtree(depth=depth, index=random_point, subtree=new_subtree)

    return individual


import copy
from typing import Callable

def initialise_population_full_method(adam, n : int, node_tree_create_func : Callable = None, island=None):
    """
    Classic random initialisation function to create a pool of `n` programs from a starting GeneticProgram `adam`.
    This method grows a population of programs each with node trees that extend to `max_depth` for the whole tree.

    Args:
        adam (GeneticProgram): A genetic program already initialised with a node tree.
        n (int): Population size.
        island (Island): Needed to wrap to `create_node` and `create_genetic_program` methods.

    Returns:
        list: Population members.
    """
    population = list()

    for i in range(n - 1):
        if not node_tree_create_func is None:
            node_tree = node_tree_create_func()
            t_func = node_tree_create_func
        else:
            node_tree = adam.tree_generator(growth_mode='full', genetic_program=adam)
            t_func = adam.tree_generator

        eve = island.create_genetic_program(adam.fitness_function,
                                            node_tree=node_tree,
                                            operators=adam.operators,
                                            terminals=adam.terminals,
                                            max_depth=adam.max_depth,
                                            min_depth=adam.min_depth,
                                            growth_mode='full',
                                            terminal_prob=adam.terminal_prob,
                                            tree_generator=t_func,
                                            species_type=adam.species_type,
                                            program_properties=copy.deepcopy(adam.get_properties()))
        population.append(eve)

    population.append(adam)

    return population

def initialise_population_grow_method(adam, n : int, node_tree_create_func : Callable = None, island=None):
    """
    Classic random initialisation function to create a pool of `n` programs from a starting GeneticProgram `adam`.
    Node trees are grown from the root but not necessarily to `max_depth`.

    Args:
        adam (GeneticProgram): A genetic program already initialised with a node tree.
        n (int): Population size.
        island (Island): Needed to wrap to `create_node` and `create_genetic_program` methods.

    Returns:
        list: Population members.
    """
    population = list()

    for i in range(n - 1):
        if not node_tree_create_func is None:
            node_tree = node_tree_create_func()
            t_func = node_tree_create_func
        else:
            node_tree = adam.tree_generator(growth_mode='grow', genetic_program=adam)
            t_func = adam.tree_generator

        eve = island.create_genetic_program(adam.fitness_function,
                                            node_tree=node_tree,
                                            operators=adam.operators,
                                            terminals=adam.terminals,
                                            max_depth=adam.max_depth,
                                            min_depth=adam.min_depth,
                                            growth_mode='grow',
                                            terminal_prob=adam.terminal_prob,
                                            tree_generator=t_func,
                                            species_type=adam.species_type,
                                            program_properties=copy.deepcopy(adam.get_properties()))
        population.append(eve)

    population.append(adam)

    return population


from datetime import datetime
import unittest
from natural_selection.genetic_algorithms import Gene, Chromosome, Individual
from natural_selection import Island
from natural_selection.genetic_algorithms.utils.random_functions import random_int
from natural_selection.genetic_algorithms.operators.crossover import crossover_two_n_point

import time

def fitness(individual, island, x, y):
    time.sleep(1)
    return (individual.chromosome[0].value * x) + (individual.chromosome[1].value *y)

class TestSimpleIsland(unittest.TestCase):

    def setUp(self) -> None:
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])
        self.ind = Individual(fitness, name="Adam", chromosome=gen)
        self.ind.add_new_property('some_property', 10)

    def test_raise_value(self):
        with self.assertRaises(ValueError):
            life = Island({'x': 0.6, 'y': 0.2}, crossover_function=crossover_two_n_point, core_count=1.1, verbose=False)

        with self.assertRaises(ValueError):
            life = Island({'x': 0.6, 'y': 0.2}, crossover_function=crossover_two_n_point, core_count=-0.5, verbose=False)

        with self.assertRaises(ValueError):
            life = Island({'x': 0.6, 'y': 0.2}, crossover_function=crossover_two_n_point, core_count="-0.5", verbose=False)

    def test_import_migrants(self):
        life = Island({'x': 0.6, 'y': 0.2}, crossover_function=crossover_two_n_point, core_count=1, verbose=False)
        life.initialise(self.ind, population_size=2)
        g_1 = Gene(name="first", value=1, gene_max=25, gene_min=1, randomise_function=random_int)
        g_2 = Gene(name="second", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_3 = Gene(name="third", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_4 = Gene(name="fourth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_5 = Gene(name="fifth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        g_6 = Gene(name="sixth", value=1, gene_max=100, gene_min=1, randomise_function=random_int)
        gen = Chromosome([g_1, g_2, g_3, g_4, g_5, g_6])
        migrants = [life.create_individual(fitness, name=f"alien_{i}", chromosome=gen) for i in range(10)]

        for m in migrants:
            m.chromosome.randomise_all_genes()

        start = datetime.utcnow()
        life.import_migrants(migrants)
        seq_end = datetime.utcnow() - start

        life = Island({'x': 0.6, 'y': 0.2}, crossover_function=crossover_two_n_point, core_count=-1, verbose=False)

        start = datetime.utcnow()
        life.import_migrants(migrants)
        para_end = datetime.utcnow() - start

        print(f'Migrants, Seq: {seq_end} Para: {para_end}')
        self.assertLess(para_end, seq_end)


    def test_evolve(self):
        life = Island({'x': 0.6, 'y': 0.2}, crossover_function=crossover_two_n_point, core_count=1, verbose=False)

        start = datetime.utcnow()
        life.initialise(self.ind, population_size=10)
        seq_end = datetime.utcnow() - start
        start = datetime.utcnow()
        life.evolve(parent_selection_params={'n':8}, crossover_probability=1.0)
        seq_end_2 = datetime.utcnow() - start

        life = Island({'x': 0.6, 'y': 0.2}, crossover_function=crossover_two_n_point, core_count=-1, verbose=False)

        start = datetime.utcnow()
        life.initialise(self.ind, population_size=10)
        para_end = datetime.utcnow() - start
        start = datetime.utcnow()
        life.evolve(parent_selection_params={'n': 8}, crossover_probability=1.0)
        para_end_2 = datetime.utcnow() - start

        self.assertLess(para_end, seq_end)
        self.assertLess(para_end_2, seq_end_2)

        print(f'Init, Seq: {seq_end} Para: {para_end}')
        print(f'Evo, Seq: {seq_end_2} Para: {para_end_2}')







import unittest
import natural_selection.genetic_programs.node_operators as ops
from natural_selection.genetic_programs import GeneticProgram


class TestGeneticProgramMemoisation(unittest.TestCase):

    def test_is_deterministic(self):

        operators = [ops.OperatorAdd]
        terminals = ['x', 2]

        gp_symb_first = GeneticProgram(operators=operators,
                                 terminals=terminals,
                                 max_depth=3, min_depth=2, growth_mode='grow',
                                 is_deterministic=True)

        gp_symb_second = GeneticProgram(operators=operators,
                                       terminals=terminals,
                                       max_depth=3, min_depth=2, growth_mode='grow',
                                       is_deterministic=True)

        result = gp_symb_first(x=3)

        result = gp_symb_second(x=3)

        self.assertTrue(True)






import unittest
from natural_selection.genetic_programs.node_operators import OperatorAdd, OperatorSub, OperatorMul, OperatorDiv, Operator
from natural_selection.genetic_programs.utils import GeneticProgramError
from natural_selection.genetic_programs import Node, GeneticProgram
import random

class TestGeneticProgram(unittest.TestCase):

    def test_random_generate(self):
        operators = [OperatorAdd, OperatorSub]
        terminals = ['X', 2]

        gp_full = GeneticProgram(operators=operators, terminals=terminals, max_depth=5, min_depth=1, growth_mode='full')


        # Must be of depth 5 because
        depth = gp_full.node_tree.depth()

        self.assertEqual(depth, 5)

        function_value = gp_full(X=3)

        self.assertIsInstance(function_value, int)

        gp_grow = GeneticProgram(operators=operators, terminals=terminals, max_depth=3, min_depth=1, growth_mode='grow')

        depth = gp_grow.node_tree.depth()

        self.assertLessEqual(depth, 3)

        function_value = gp_grow(X=3)

        self.assertIsInstance(function_value, int)

        gp_grow = GeneticProgram(operators=operators, terminals=terminals, max_depth=4, min_depth=3, growth_mode='grow')

        depth = gp_grow.node_tree.depth()

        self.assertLessEqual(depth, 4)
        self.assertGreaterEqual(depth, 3)

        function_value = gp_grow(X=3)

        self.assertIsInstance(function_value, int)

        gp_grow = GeneticProgram(operators=operators, terminals=terminals, max_depth=4, min_depth=4, growth_mode='grow')

        depth = gp_grow.node_tree.depth()

        self.assertEqual(depth, 4)

    def test_brute_check_depth(self):
        operators = [OperatorAdd, OperatorSub]
        terminals = ['X', 2]

        # Full gen
        for i in range(50):
            rand_min = random.randint(1, 5)
            rand_max = random.randint(rand_min+1, rand_min+5)
            gp = GeneticProgram(operators=operators, terminals=terminals, max_depth=rand_max, min_depth=rand_min, growth_mode='full')

            self.assertEqual(gp.node_tree.depth(), rand_max)

        # Grow gen
        for i in range(50):
            rand_min = random.randint(1, 5)
            rand_max = random.randint(rand_min+1, rand_min+5)

            gp = GeneticProgram(operators=operators, terminals=terminals, max_depth=rand_max, min_depth=rand_min,
                                     growth_mode='grow')

            depth = gp.node_tree.depth()

            self.assertLessEqual(depth, rand_max)
            self.assertGreaterEqual(depth, rand_min)

        for i in range(50):
            rand_depth = random.randint(2, 10)

            gp = GeneticProgram(operators=operators, terminals=terminals, max_depth=rand_depth, min_depth=rand_depth,
                                growth_mode='grow')

            self.assertEqual(gp.node_tree.depth(), rand_depth)

        for i in range(50):

            gp = GeneticProgram(operators=operators, terminals=terminals, max_depth=1, min_depth=1,
                                growth_mode='grow')

            self.assertEqual(gp.node_tree.depth(), 1)



    def test_evaluate(self):
        def func(args):
            return abs(args[0] - args[1]) / (args[0] + args[1])

        custom = Operator(operator_label='custom', function=func, min_arity=2, max_arity=2)

        n1 = Node(label='X', is_terminal=True)
        n2 = Node(label='Y', is_terminal=True)

        n = Node(arity=2, operator=custom, children=[n1, n2])

        def fitness_func(program, island, X, Y):
            return program(X=X, Y=Y)


        gp = GeneticProgram(node_tree=n, fitness_function=fitness_func)

        fitness = gp.evaluate({'X':15, 'Y':26})
        expected_fitness = func([15,26])

        self.assertEqual(fitness, expected_fitness)

        # Test no fitness function given
        gp = GeneticProgram(node_tree=n)

        fitness = gp.evaluate({'X': 16, 'Y': 24})
        expected_fitness = func([16, 24])

        self.assertEqual(fitness, expected_fitness)

        node = Node(arity=1, operator=None, terminal_value=77, is_terminal=True)
        gp = GeneticProgram(node_tree=node)

        fitness = gp.evaluate()

        self.assertEqual(fitness, 77)


    def test_empty_arg_call(self):
        node = Node(arity=1,operator=None,terminal_value=39, is_terminal=True)
        gp = GeneticProgram(node_tree=node)

        return_value = gp()

        self.assertEqual(return_value, 39)

        node = Node(arity=1, label='X', operator=None, is_terminal=True)

        gp.node_tree = node

        return_value = gp()

        self.assertIsNone(return_value)

        return_value = gp(X=42)

        self.assertEqual(return_value, 42)

        pass